module Core where

import Data.List
import Data.Maybe
import System.Process
import System.Posix.Process
import GHC.Paths
import System.Exit
import System.FilePath
import System.Directory
import Data.Binary
import System.IO
import System.Info
import Control.Exception

import Manatee.Utils.Time

-- | Main name for generate binary entry.
mainName :: String
mainName = "Manatee"

-- | Main file for recompile.
mainFile :: IO FilePath
mainFile = do
  curDir <- getCurrentDirectory
  return $ curDir </> mainName ++ ".hs"

-- | Cache directory for contain last correct file.
cacheDir :: FilePath
cacheDir = "/test/Download/cache/"
-- cacheDir = "/home/andy/cache/"

-- | Binary entry name.
cacheExec :: FilePath
cacheExec = cacheDir </> mainName ++ "-" ++ os ++ "-" ++ arch

-- | Save state for through recompile.
saveState :: Binary a => a -> IO ()
saveState state = do
  -- Get path to save state.
  statePath <- genStatePath

  -- Encode state to file.
  encodeFile statePath . Just $ state

-- | Restore state after recompile.
restoreState :: Binary a => a -> IO a
restoreState d = do  
  -- Get state path.
  statePath <- genStatePath

  exists <- doesFileExist statePath
  if exists
     -- Return save state when state path exist.
     then do
       state <- decodeFile statePath
       return $ fromMaybe d state
     -- Otherwise, return default state.
     else return d

-- | Get state path with PID.
genStatePath :: IO FilePath
genStatePath = do
  -- Get PID.
  pidString <- getPIDString

  -- Make state file store in TEMP directory of OS.
  tempDir   <- getTemporaryDirectory
  let statePath = tempDir </> pidString ++ ".state"

  return statePath

-- | Recompile all program files.
-- Return Nothing when recompile successful,    
-- otherwise, return compile error information.    
recompile :: IO (Maybe String)
recompile = do
  createDirectoryIfMissing True cacheDir

  -- Compile occurs in here
  let errFile = cacheDir </> "errors.log"
  result <- bracket (openFile errFile WriteMode) hClose $ \errHandle -> do
              -- Get directory that contain current binary entry.
              currentDir <- getCurrentDirectory
              -- Get main file.
              mFile <- mainFile
              -- Config compile options.
              let ghcOpts = ["-v0"
                            ,"-i" ++ currentDir
                            ,"-outputdir", cacheDir
                            ,"--make", mFile
                            ,"-o", cacheExec]
              -- Run compile process.
              ghcProc <- runProcess ghc ghcOpts (Just cacheDir) Nothing
                        Nothing Nothing (Just errHandle)
              waitForProcess ghcProc

  -- Display a helpful little status message
  putStrLn $ if result /= ExitSuccess
                then "Error occurred while loading configuration file."
                else "Program reconfiguration successful."

  errorsExist <- doesFileExist errFile
  if errorsExist
     then do 
       errors <- readFile errFile
       if errors == ""
          then return Nothing
          -- Return compile error when compile failed.
          else return . Just $ errors
     -- Otherwise return Nothing.
     else return Nothing
          
-- | Relaunch.
-- Relaunch new binary entry when recompile successful,
-- otherwise, output compile error.  
relaunch :: Binary a => a -> IO ()
relaunch state = do
  -- Show start time.
  startTime <- getCurrentLocalTime
  putStrLn $ show startTime ++ ":> Relaunch ......"

  -- Compile.
  info <- recompile

  -- Show end time.
  endTime <- getCurrentLocalTime
  putStrLn $ show endTime ++ ":> Relaunch end."

  case info of
    -- Show compile error when relaunch failed.
    Just i  -> putStrLn i
    -- Otherwise, save state and execute new binary entry. 
    Nothing -> do
              saveState state
              executeFile cacheExec False [] Nothing
      
-- | Get PID string.
getPIDString :: IO String
getPIDString = fmap show getProcessID
