module Manatee.Data.ListZipper where

import Data.Maybe

import Manatee.Utils.Generic

import qualified Data.List as List

data ListZipper a = ListZipper ![a] ![a]

instance Show a => Show (ListZipper a) where
    show (ListZipper ls rs) = show (reverse ls) ++ show rs

instance Functor ListZipper where
    fmap f (ListZipper ls rs) = ListZipper (map f ls) (map f rs)

-- | Create an empty ListZipper. 
empty :: ListZipper a
empty = ListZipper [] []

-- | Create a ListZipper with a single element.
singletion :: a -> ListZipper a
singletion a = ListZipper [] [a]

fromList, fromListEnd :: [a] -> ListZipper a
-- | Create a ListZipper from list, and focus focus to first element in list.
fromList = ListZipper []
-- | Create a ListZipper from list end, and focus focus to end element in list.
fromListEnd as = ListZipper (reverse as) []

-- | To List.
toList :: ListZipper a -> [a]
toList (ListZipper ls rs) = reverse ls ++ rs

atStart, atEnd, isEmpty :: ListZipper a -> Bool
-- | Whether is first element.
atStart (ListZipper [] _ ) = True
atStart _                  = False
-- | Whether is last element.
atEnd   (ListZipper _  []) = True
atEnd   _                  = False
-- | Whether is empty.
isEmpty (ListZipper [] []) = True
isEmpty _                  = False

getCurrent, getLeft, getRight, getFirst, getLast :: ListZipper a -> Maybe a
-- | Get current node, return Nothing if haven't found anything in current node.
-- This function won't change current focus node.
getCurrent list = get list $ currentIndex list
-- | Get left node, return Nothing if haven't found anything in left node.
-- This function won't change current focus node.
getLeft    list = get list $ leftIndex list
-- | Get right node, return Nothing if haven't found anything in right node.
-- This function won't change current focus node.
getRight   list = get list $ rightIndex list
-- | Get first node, return Nothing if haven't found anything in first node.
-- This function won't change current focus node.
getFirst   list = get list $ firstIndex list
-- | Get last node, return Nothing if haven't found anything in last node.
-- This function won't change current focus node.
getLast    list = get list $ lastIndex list

getLeftCircular, getRightCircular :: ListZipper a -> Maybe a
-- | Get left node, return Nothing if haven't found anything in left node.
-- If reach first node, circular to end node.
-- This function won't change current focus node.
getLeftCircular  list
    | prevIndex < firstIndex list 
        = get list $ lastIndex list
    | otherwise 
        = get list prevIndex
    where
      prevIndex = leftIndex list
-- | Get right node, return Nothing if haven't found anything in right node.
-- If reach last node, circular to first node.
-- This function won't change current focus node.
getRightCircular list
    | nextIndex > lastIndex list 
        = get list $ firstIndex list
    | otherwise 
        = get list nextIndex
    where
      nextIndex = rightIndex list

-- | Get node with given index.
-- This function won't change current focus node.
get :: ListZipper a -> Int -> Maybe a
get = maybeIndex . toList 

focusLeft, focusRight, focusFirst, focusLast :: ListZipper a -> Maybe (ListZipper a)
-- | Focus to left node.
focusLeft  (ListZipper (a:ls) rs) = Just $ ListZipper ls (a:rs)
focusLeft  _                      = Nothing
-- | Focus to right node.
focusRight (ListZipper ls (a:rs)) = Just $ ListZipper (a:ls) rs
focusRight _                      = Nothing
-- | Focus to first node.
focusFirst list                   = focus (firstIndex list) list
-- | Focus to last node.
focusLast  list                   = focus (lastIndex list) list

-- | Focus focus to given index node.
focus :: Int -> ListZipper a -> Maybe (ListZipper a)
focus n list@(ListZipper _ _) 
     | n < 0 || n >= lenList 
         = Nothing
     | n == currIndex        
         = Just list
     | n < currIndex         
         = applyMaybeT (focus n) leftList
     | otherwise             
         = applyMaybeT (focus n) rightList
     where
       lenList   = Manatee.Data.ListZipper.length list
       currIndex = currentIndex list
       leftList  = focusLeft list
       rightList = focusRight list

-- | Focus node.
focusNode :: Eq a => a -> ListZipper a -> Maybe (ListZipper a)
focusNode a list = 
    applyMaybeT (\i -> focus i list) index
    where
      index = List.findIndex (a ==) (toList list)

insertLeft, insertRight, insertFirst, insertLast :: a -> ListZipper a -> ListZipper a
-- | Insert node to right side of current node, and keep current focus.
insertLeft  a (ListZipper ls rs)     = ListZipper (a:ls) rs
-- | Insert node to left side of current node, and keep current focus.
-- Except it will focus first element when list is empty.
insertRight a (ListZipper ls (r:rs)) = ListZipper ls (r:a:rs)
insertRight a (ListZipper ls _)      = ListZipper ls [a]
-- | Insert node to first end, and keep current focus.
insertFirst a (ListZipper (l:ls) rs) = ListZipper ((l:ls) ++ [a]) rs
insertFirst a (ListZipper _ rs)      = ListZipper [a] rs
-- | Insert node to last end, and keep current focus.
insertLast  a (ListZipper ls (r:rs)) = ListZipper ls ((r:rs) ++ [a])
insertLast  a (ListZipper ls _)      = ListZipper ls [a]

delete, deleteLeft, deleteRight, deleteFirst, deleteLast, deleteOthers :: ListZipper a -> Maybe (ListZipper a)
-- | Delete current node, and focus focus right node.
delete       (ListZipper ls (_:rs))   = Just $ ListZipper ls rs
delete       _                        = Nothing
-- | Delete left node, and keep current focus.
deleteLeft   (ListZipper (_:ls) rs)   = Just $ ListZipper ls rs
deleteLeft   _                        = Nothing 
-- | Delete right node, and keep current focus.
deleteRight  (ListZipper ls (r:_:rs)) = Just $ ListZipper ls (r:rs)
deleteRight  _                        = Nothing
-- | Delete first node, and keep current focus.
deleteFirst  (ListZipper (l:ls) rs)   = Just $ ListZipper (init (l:ls)) rs
deleteFirst  (ListZipper [] (_:rs))   = Just $ ListZipper [] rs
deleteFirst  _                        = Nothing
-- | Delete last node, and keep current focus.
deleteLast   (ListZipper ls (r:rs))   = Just $ ListZipper ls (init (r:rs))
deleteLast   (ListZipper (_:ls) rs)   = Just $ ListZipper ls rs
deleteLast   _                        = Nothing
-- | Delete other node.
deleteOthers (ListZipper _ (r:_))     = Just $ ListZipper [] [r]
deleteOthers _                        = Nothing

swapLeft, swapRight :: ListZipper a -> Maybe (ListZipper a)
-- | Swap with left node, and focus swap node.
swapLeft  (ListZipper (l:ls) (r:rs)) = Just $ ListZipper (r:ls) (l:rs)
swapLeft  (ListZipper (l:ls) _)      = Just $ ListZipper ls [l]
swapLeft  _                          = Nothing
-- | Swap with right node, and focus swap node.
swapRight (ListZipper ls (r:x:rs))   = Just $ ListZipper ls (x:r:rs)
swapRight _                          = Nothing

-- | Replace current node.
replace :: a -> ListZipper a -> Maybe (ListZipper a)
replace a (ListZipper ls (_:rs)) = Just $ ListZipper ls (a:rs)
replace _ _                      = Nothing

length, currentIndex, leftIndex, rightIndex, firstIndex, lastIndex :: ListZipper a -> Int
-- | Get length.
length       (ListZipper ls rs) = Prelude.length ls + Prelude.length rs 
-- | Get current inde.
currentIndex (ListZipper ls _)  = Prelude.length ls
-- | Get left index.
leftIndex                       = pred . currentIndex 
-- | Get right index.
rightIndex                      = succ . currentIndex
-- | Get first index.
firstIndex   _                  = 0
-- | Get last index.
lastIndex                       = pred . Manatee.Data.ListZipper.length
