{-# LANGUAGE ExistentialQuantification, RankNTypes #-}
module Manatee.IORefObject where

import Data.IORef
import Control.Monad

import Graphics.UI.Gtk

import Manatee.Utils.Generic hiding (swap)
import Manatee.Utils.Seq 
import {-# source #-} Manatee.UI.View
import {-# source #-} Manatee.UI.ViewNode
import {-# source #-} Manatee.UI.Page
import {-# source #-} Manatee.UI.TabGroup
import {-# source #-} Manatee.PageView.PageView
import {-# source #-} Manatee.UI.Interactivebar

data IORefObject = 
    IORefObject {ioRefObjectRootWindow          :: Window
                ,ioRefObjectViewList            :: IORef ViewList
                ,ioRefObjectViewNodeList        :: IORef ViewNodeList
                ,ioRefObjectPageList            :: IORef PageList
                ,ioRefObjectTerminalCounter     :: IORef Int}

-- | Create new IORef object.
ioRefObjectNew :: IO IORefObject
ioRefObjectNew = do
  -- View list.
  viewList <- viewListNew

  -- View node list.
  viewNodeList <- viewNodeListNew

  -- Page list.
  pageList <- pageListNew

  -- Root window.
  rootWindow <- windowNew
  windowFullscreen rootWindow

  -- Terminal counter. 
  terminalCounter <- newIORef 0

  -- Create new ioRefObject.
  let ioRefObject = IORefObject rootWindow viewList viewNodeList pageList terminalCounter 

  return ioRefObject

-- | Get focus view.
ioRefObjectGetFocusView :: IORefObject -> IO (Maybe View) 
ioRefObjectGetFocusView = viewListGetFocusView . ioRefObjectViewList

-- | Get focus page.
ioRefObjectGetFocusPage :: IORefObject -> IO (Maybe Page)
ioRefObjectGetFocusPage = 
    viewApplyFocus' viewGetCurrentPage 

-- | Get focus tab group.
ioRefObjectGetFocusTabGroup :: IORefObject -> IO (Maybe TabGroup)
ioRefObjectGetFocusTabGroup = 
    (<<<=) (applyMaybe viewTabGroup) ioRefObjectGetFocusView  

-- | Get focus interactivebar.
ioRefObjectGetFocusInteractivebar :: IORefObject -> IO (Maybe Interactivebar)
ioRefObjectGetFocusInteractivebar ioRefObject =
    applyMaybe_ (pageViewUIFrameInteractivebar . pageApplyBuffer pageViewUIFrame
                ) =<< ioRefObjectGetFocusPage ioRefObject

-- | Add page.
ioRefObjectAddPage :: Page -> IORefObject -> IO ()
ioRefObjectAddPage page ioRefObject = do
  let pageList = ioRefObjectPageList ioRefObject
  
  -- Add page to page list.
  pageListAddPage pageList page
  
  -- Synchronize page list with all view.
  applyMaybeTM_ (\view -> do
                   pageId <- viewGetCurrentPageSequenceId view
                   let (position, pId) = case pageId of
                                           Just i  -> (PRight, i)
                                           Nothing -> (PStart, 0)
                   viewListApplyView (\view -> do
                                        -- Get clone page.
                                        clonePage <- pageClone page

                                        -- Add clone page to tabGroup.
                                        tabGroupAddPage (viewTabGroup view) position pId clonePage

                                        -- Remove page when page box destroy.
                                        pageBox clonePage `onDestroy` pageRemove clonePage ioRefObject >> return ()
                                     ) ioRefObject
                ) =<< ioRefObjectGetFocusView ioRefObject

-- | Remove page.
ioRefObjectRemovePage :: Page -> IORefObject -> IO ()
ioRefObjectRemovePage page ioRefObject = do
  let pageList = ioRefObjectPageList ioRefObject
  removePage <- pageListGetPage pageList $ pageId page
  applyMaybeTM_ (\rp -> do
                   -- Remove page that have same page id from page list.
                   pageListRemovePage pageList rp

                   -- Remove page that have same page id from all views.
                   viewListApplyView (\view -> tabGroupRemovePage (viewTabGroup view) page) ioRefObject
                ) removePage
  
-- | Set page name.
ioRefObjectSetPageName :: Page -> String -> IORefObject -> IO ()
ioRefObjectSetPageName page name ioRefObject = do
  -- Set page name.
  pageSetName page name
  
  -- Synchronize with view list.
  viewListApplyView (\view -> tabGroupSetPageName (viewTabGroup view) page name) ioRefObject 

-- | Move page.
ioRefObjectMovePage :: Page -> Int -> IORefObject -> IO ()
ioRefObjectMovePage page id =
    viewListApplyView (\view -> do
                         -- Get tabGroup.
                         let tabGroup = viewTabGroup view
                         applyMaybeTM_ (\i -> 
                                        applyMaybeTM_ (\cp -> do
                                                         -- Get notebook and child.
                                                         let notebook = tabGroupNotebook tabGroup
                                                             box = pageBox cp

                                                         -- Reorder child in notebook.
                                                         notebookReorderChild notebook box id

                                                         -- Synchronize sequence id in TabGroup.
                                                         modifyIORef (tabGroupPageIdSequence tabGroup) (swap i id)
                                                      ) =<< tabGroupGetPageSequncePage tabGroup i
                                       ) =<<  tabGroupGetPageSequnceId tabGroup page) 
  
    