{-# LANGUAGE ScopedTypeVariables #-}
module Manatee.PageView.DiredBuffer where

import System.IO
import System.Posix.Types
import Control.Monad
import System.Gnome.VFS
import Data.IORef
import System.Locale
import System.Time
import Data.Maybe
import System.FilePath

import Graphics.UI.Gtk.General.Enums
import Graphics.UI.Gtk.ModelView.TreeSortable

import Manatee.Utils.Generic
import Manatee.Utils.Gtk.Gtk

import qualified System.Gnome.VFS as VFS

data DiredBuffer = 
    DiredBuffer {diredBufferCurrentDirectory    :: FilePath
                ,diredBufferFileInfoOptions     :: [(FileInfoOption, SortColumnId)]
                ,diredBufferFileInfos           :: [DiredFileInfo]
                ,diredBufferSortStatus          :: (FileInfoOption, SortType)}

data DiredFileInfo =
    DiredFileInfo {fileInfoNameDescrible    :: (String,    String)
                  ,fileInfoTypeDescrible    :: (FileType,  String)
                  ,fileInfoSizeDescrible    :: (FileSize,  String)
                  ,fileInfoItemDescrible    :: (Int,       String)
                  ,fileInfoMimeDescrible    :: (String,    String)
                  ,fileInfoMtimeDescrible   :: (EpochTime, String)
                  ,fileInfoAtimeDescrible   :: (EpochTime, String)
                  ,fileInfoCtimeDescrible   :: (EpochTime, String)}
    deriving (Show)

data FileInfoOption = FIName | FISize | FIItem | FIMime | FIMTime | FIATime | FICTime
                    deriving (Eq, Show, Read)

class DiredFileInfoClass a where
    getColumnTitle      :: a -> String -- get title for treeColumn 
    getCellText         :: a -> DiredFileInfo -> String -- get text for cell
    getCellXAlign       :: a -> Float                  -- get x align for cell
    compareRow          :: a -> DiredFileInfo -> DiredFileInfo -> IO Ordering -- compare row for treeView

instance DiredFileInfoClass FileInfoOption where
      getColumnTitle FIName  = "File name"
      getColumnTitle FISize  = "Size"
      getColumnTitle FIItem  = "Item"
      getColumnTitle FIMime  = "Type"
      getColumnTitle FIMTime = "Modification time"
      getColumnTitle FICTime = "Create time"
      getColumnTitle FIATime = "Access time"

      getCellText FIName  info = snd $ fileInfoNameDescrible info
      getCellText FISize  info = snd $ fileInfoSizeDescrible info
      getCellText FIItem  info = snd $ fileInfoItemDescrible info
      getCellText FIMime  info = snd $ fileInfoMimeDescrible info
      getCellText FIMTime info = snd $ fileInfoMtimeDescrible info
      getCellText FICTime info = snd $ fileInfoCtimeDescrible info
      getCellText FIATime info = snd $ fileInfoAtimeDescrible info

      getCellXAlign FIName  = 0.0
      getCellXAlign FISize  = 1.0
      getCellXAlign FIItem  = 1.0
      getCellXAlign FIMime  = 1.0
      getCellXAlign FIMTime = 0.0
      getCellXAlign FICTime = 0.0
      getCellXAlign FIATime = 0.0

      compareRow FIName  row1 row2 = return $ diredFileInfoNameCompare row1 row2
      compareRow FISize  row1 row2 = return $ compare (fst $ fileInfoSizeDescrible  row1) (fst $ fileInfoSizeDescrible  row2)
      compareRow FIItem  row1 row2 = return $ compare (fst $ fileInfoItemDescrible  row1) (fst $ fileInfoItemDescrible  row2)
      compareRow FIMime  row1 row2 = return $ compare (fst $ fileInfoMimeDescrible  row1) (fst $ fileInfoMimeDescrible  row2)
      compareRow FIMTime row1 row2 = return $ compare (fst $ fileInfoMtimeDescrible row1) (fst $ fileInfoMtimeDescrible row2)
      compareRow FICTime row1 row2 = return $ compare (fst $ fileInfoCtimeDescrible row1) (fst $ fileInfoCtimeDescrible row2)
      compareRow FIATime row1 row2 = return $ compare (fst $ fileInfoAtimeDescrible row1) (fst $ fileInfoAtimeDescrible row2)

-- | New.
diredBufferNew :: FilePath -> IO DiredBuffer
diredBufferNew dir = 
    handleVFSError $ do
      initVFS
      let infoOptions = (pairPred [FIName, FISize, FIItem, FIMime, FIMTime, FIATime, FICTime])
      fileInfos <- diredBufferGenerateFileInfos dir
      let sortStatus  = (FIName, SortAscending)
      return $ DiredBuffer dir infoOptions fileInfos sortStatus

-- | Empty buffer.
diredBufferEmpty :: DiredBuffer
diredBufferEmpty = DiredBuffer "" [] [] (FIName, SortAscending)

-- | Load.
diredBufferLoad :: IORef DiredBuffer -> FilePath -> IO ()
diredBufferLoad buffer dir = 
  handleVFSError $ do
    fileInfos <- diredBufferGenerateFileInfos dir
    modifyIORef buffer (\x -> x {diredBufferCurrentDirectory = dir
                                ,diredBufferFileInfos        = fileInfos})

-- | Get file infos.
diredBufferGenerateFileInfos :: FilePath -> IO [DiredFileInfo]
diredBufferGenerateFileInfos dir = do
  infos <- directoryListLoad dir []
  let fileInfos = filterDotDirectoryApply (fromMaybe "" . fileInfoName) infos
  mapM (\fi -> do
          -- Get type.
          typeDes <- do
            let fType = fromMaybe FileTypeUnknown (fileInfoType fi)
                title = show fType
            return (fType, title)

          -- Get name.
          nameDes <- do
            let name  = fromMaybe "" (fileInfoName fi)
                title = addTrailingPathSeparatorWithFileType name (fst typeDes)
            return (name, title)

          -- Get size.
          sizeDes <- do
            let size  = fromMaybe 0 (fileInfoSize fi)
                title = VFS.formatFileSizeForDisplay size
            return (size, title)

          -- Get item number.
          itemDes <- do
            let name = (fromMaybe "" (fileInfoName fi))
            item <- getFileItems (dir </> name)
            let title = show item
            return (item, title)

          -- Get MIME.
          mimeDes <- do
            mime <- getFileMIMETypeFast (fromMaybe "" (fileInfoName fi))
            return (mime, mime)
                              
          -- Generate time.
          let generateTime f = do
                let time = fromMaybe 0 (f fi)
                title <- do
                  calTime <- toCalendarTime (TOD (fromIntegral $ fromEnum time) 0)
                  return (formatCalendarTime defaultTimeLocale "%D %T" calTime)
                return (time, title)

          -- Get time information.
          mtimeDes <- generateTime fileInfoMTime
          ctimeDes <- generateTime fileInfoCTime
          atimeDes <- generateTime fileInfoATime

          return $ DiredFileInfo nameDes typeDes sizeDes itemDes mimeDes mtimeDes atimeDes ctimeDes 
       ) fileInfos
  
-- | Compare file name.
diredFileInfoNameCompare :: DiredFileInfo -> DiredFileInfo -> Ordering
diredFileInfoNameCompare row1 row2 = 
  compareFile (fName1, fType1) (fName2, fType2)
  -- Get file type and name.
  where fType1 = fst $ fileInfoTypeDescrible row1
        fType2 = fst $ fileInfoTypeDescrible row2
        fName1 = fst $ fileInfoNameDescrible row1
        fName2 = fst $ fileInfoNameDescrible row2
