{-# LANGUAGE DeriveDataTypeable, ScopedTypeVariables #-}
module Manatee.PageView.DiredView where

import Data.IORef
import Data.Binary
import Data.List
import Control.Monad
import Data.Typeable
import Control.Arrow
import System.FilePath
import Data.Maybe

import Graphics.UI.Gtk hiding (Statusbar, statusbarNew, get)
import Graphics.UI.Gtk.ModelView
import Graphics.UI.Gtk.ModelView.TreeView
import Graphics.UI.Gtk.General.Enums

import Manatee.PageView.DiredBuffer
import Manatee.Widget.Statusbar
import Manatee.Utils.System
import Manatee.IORefObject
import Manatee.Utils.Generic
import Manatee.PageView.PageView
import Manatee.UI.Page
import Manatee.UI.View
import Manatee.UI.TabGroup
import Manatee.Utils.Gtk.Gtk
import Manatee.Keymap.Keymap
import Manatee.Data.TypeId

import qualified Data.Map as M

data DiredView = 
    DiredView {diredViewName                :: IORef String
              ,diredViewUIFrame             :: PageViewUIFrame
              ,diredViewBuffer              :: IORef DiredBuffer
              ,diredViewTreeView            :: TreeView
              ,diredViewListStore           :: ListStore DiredFileInfo
              ,diredViewSortModel           :: TypedTreeModelSort DiredFileInfo
              }
    deriving Typeable

data DiredViewAction = DiredViewAction (DiredView -> IORefObject -> IO ())

instance ActionExecute DiredViewAction where
    execute (DiredViewAction a) = diredViewApplyAction a

instance PageView DiredView where
    pageViewGetName = diredViewGetName

    pageViewSetName = diredViewSetName 

    pageViewUIFrame = diredViewUIFrame

    pageViewClone = diredViewClone

    pageViewFocus a = do
      let treeView = diredViewTreeView a
      widgetGrabFocus treeView
      treeViewFocus treeView

    pageViewScrollLineUp = diredViewScrollLineUp

    pageViewScrollLineDown = diredViewScrollLineDown

    pageViewScrollPageUp = diredViewScrollPageUp

    pageViewScrollPageDown = diredViewScrollPageDown

    pageViewBegin = diredViewBegin

    pageViewEnd = diredViewEnd

    pageViewGetKeymap _ = diredViewKeymap

    pageViewGetState a = do
      name <- diredViewGetName a
      return $ PageViewStateWrap $ DiredViewState name

diredViewStatusbar :: DiredView -> Statusbar
diredViewStatusbar = pageViewUIFrameStatusbar . diredViewUIFrame

-- | New dired view.
diredViewNew :: String -> IO DiredView
diredViewNew name = 
  diredViewNewInternal name Nothing Nothing

-- | Clone dired view.
diredViewClone :: DiredView -> IO DiredView
diredViewClone dv = do
  buffer <- diredViewGetBuffer dv
  name <- diredViewGetName dv
  diredViewNewInternal name (Just buffer) (Just $ diredViewStatusbar dv)

-- | Internal new function.
diredViewNewInternal :: String -> Maybe DiredBuffer -> Maybe Statusbar -> IO DiredView
diredViewNewInternal name bufferClone statusbarClone = do 
  -- Create UI frame.
  uiFrame <- pageViewUIFrameNew statusbarClone
  let scrolledWindow = pageViewUIFrameScrolledWindow uiFrame

  -- Name.
  newName <- newIORef name

  -- View buffer.
  buffer <- case bufferClone of
              Just n  -> newIORef n 
              Nothing -> newIORef =<< diredBufferNew name
                            
  -- Tree view.
  treeView <- treeViewNew
  treeViewSetGridLines treeView TreeViewGridLinesBoth
  treeViewSetEnableTreeLines treeView True
  scrolledWindow `containerAdd` treeView

  -- List store.
  listStore <- listStoreNew []

  -- Sort model.
  sortModel <- treeModelSortNewWithModel listStore

  -- Dired view.
  let diredView = DiredView newName uiFrame buffer treeView listStore sortModel 

  -- Draw view.
  diredViewDraw diredView

  return diredView

-- | Draw dired view.
diredViewDraw :: DiredView -> IO ()
diredViewDraw view = do
  -- Get value.
  let treeView = diredViewTreeView view
  buffer <- readIORef $ diredViewBuffer view
  let fileInfos = diredBufferFileInfos buffer 
      store     = diredViewListStore view
      model     = diredViewSortModel view

  -- Clear list store first.
  listStoreClear store
  -- Append to list store.
  mapM_ (listStoreAppend store) fileInfos

  -- Set tree view model.
  treeViewSetModel treeView model

  -- Remove all columns from tree view.
  treeViewRemoveColumns treeView

  -- Add column file info to tree view.
  mapM_ (diredViewAddColumn treeView store model) $ diredBufferFileInfoOptions buffer

  -- Sort column.
  let (sortOption, sortType) = diredBufferSortStatus buffer
  diredViewSortInternal view sortOption sortType

  return ()

-- | Add column.
diredViewAddColumn :: (DiredFileInfoClass t,
                      TreeViewClass self1,
                      TreeModelClass self,
                      TreeModelSortClass self,
                      TypedTreeModelClass model,
                      TreeSortableClass self) =>
                     self1
                   -> model DiredFileInfo
                   -> self
                   -> (t, SortColumnId)
                   -> IO ()
diredViewAddColumn treeView model sortModel option@(info,sortId) = do
  diredViewSetSortFunc model sortModel option

  let name = getColumnTitle info
  tvc <- treeViewAddColumnWithTitle treeView name sortId

  cell <- cellRendererTextNew
  treeViewColumnPackStart tvc cell True

  diredViewSetCellText tvc cell model sortModel info

-- | Set sort function.
diredViewSetSortFunc :: (TreeSortableClass self,
                        TypedTreeModelClass model,
                        DiredFileInfoClass a) =>
                       model DiredFileInfo
                     -> self
                     -> (a, SortColumnId)
                     -> IO ()
diredViewSetSortFunc model sortModel option = do 
  let info = fst option
      sortId = snd option
  treeSortableSetSortFunc sortModel sortId $ \iter1 iter2 -> do
    row1 <- treeModelGetRow model iter1
    row2 <- treeModelGetRow model iter2
    compareRow info row1 row2

-- | Set cell text.
diredViewSetCellText :: (CellLayoutClass self,
                        CellRendererTextClass cell,
                        TreeModelClass model,
                        TreeModelSortClass model,
                        TypedTreeModelClass model1,
                        DiredFileInfoClass a) =>
                       self
                     -> cell
                     -> model1 DiredFileInfo
                     -> model
                     -> a
                     -> IO ()
diredViewSetCellText tvc cell model sortModel info = 
    cellLayoutSetAttributeFunc tvc cell sortModel $ \iter -> do
      row <- treeModelSortGetRow model sortModel iter
      set cell [cellText   := getCellText info row
               ,cellXAlign := getCellXAlign info]

-- | Next ndoe.
diredViewNextNode :: DiredView -> IORefObject -> IO ()
diredViewNextNode view _ =
    treeViewNextNode (diredViewTreeView view)
    
-- | Previous node.
diredViewPrevNode :: DiredView -> IORefObject -> IO ()    
diredViewPrevNode view _ =
    treeViewPrevNode (diredViewTreeView view)

-- | Entry node.
diredViewEntryNode :: DiredView -> IORefObject -> IO ()
diredViewEntryNode view ioRefObject = 
    treeViewApplyCellPath 
    (diredViewTreeView view) 
    (\(childPath,_,_) -> 
         -- When path is valid value.
         unlessNull childPath $ do
           -- Convert sort path to unsort path.
           path <- treeModelSortConvertPathToChildPath (diredViewSortModel view) childPath
           -- Get file info.
           fileInfo <- listStoreGetValue (diredViewListStore view) (head path)
           -- Get file name.
           let n = fst $ fileInfoNameDescrible fileInfo 
           -- Get current directory.
           dir <- (<<<=) diredBufferCurrentDirectory diredViewGetBuffer view 
           -- Get full file path.
           let filePath = dir </> n
           -- Entry node.
           ifM (doesFileExist_ filePath)
                   -- Open file when path is file.
                   (fileOpen filePath ioRefObject)
                   (ifM (doesDirectoryExist_ filePath)
                    -- Open directory when path is directory.
                    (diredInternal filePath ioRefObject)
                    (return ())))
           
-- | Jump upper directory.
diredViewUpperDirectory :: DiredView -> IORefObject -> IO ()
diredViewUpperDirectory view ioRefObject = do
  -- Get current directory.
  dir <- (<<<=) diredBufferCurrentDirectory diredViewGetBuffer view 
  -- Get upper directory.
  let upperDir = getUpperDirectory dir 
  unless (dir == upperDir) $
         do
           -- Update page view name in page list.
           applyMaybeTM_ (\p -> 
                              applyMaybeTM_ (pageApplyBuffer (\x -> pageViewSetName x upperDir)) 
                                  =<< pageListGetPage (ioRefObjectPageList ioRefObject) (pageId p)) 
               =<< ioRefObjectGetFocusPage ioRefObject

           -- Load new directory path.
           diredBufferLoad (diredViewBuffer view) upperDir

           -- Draw again.
           diredViewDraw view

           -- Unique page name.
           pageNameUnique ioRefObject

-- | Sort by file name.
diredViewSortByName :: DiredView -> IORefObject -> IO ()
diredViewSortByName view _ = diredViewSort view FIName

-- | Sort by file size.
diredViewSortBySize :: DiredView -> IORefObject -> IO ()
diredViewSortBySize view _ = diredViewSort view FISize

-- | Sort by file item.
diredViewSortByItem :: DiredView -> IORefObject -> IO ()
diredViewSortByItem view _ = diredViewSort view FIItem

-- | Sort by file mime type.
diredViewSortByMime :: DiredView -> IORefObject -> IO ()
diredViewSortByMime view _ = diredViewSort view FIMime

-- | Sort by modified time.
diredViewSortByMTime :: DiredView -> IORefObject -> IO ()
diredViewSortByMTime view _ = diredViewSort view FIMTime

-- | Sort by access time.
diredViewSortByATime :: DiredView -> IORefObject -> IO ()
diredViewSortByATime view _ = diredViewSort view FIATime

-- | Sort by create time.
diredViewSortByCTime :: DiredView -> IORefObject -> IO ()
diredViewSortByCTime view _ = diredViewSort view FICTime

-- | Sort column.
diredViewSort :: DiredView -> FileInfoOption -> IO ()
diredViewSort view option = do
  -- Get model and options.
  let model = diredViewSortModel view 
  options <- (<<<=) diredBufferFileInfoOptions diredViewGetBuffer view 
  -- Get current sortType and columnId.
  (curSortType, _, curSortColumnId) <- treeSortableGetSortColumnId model
  applyMaybeTM_ (\id -> do
                     treeSortableSetSortColumnId model id $
                                                 if id == curSortColumnId
                                                    -- Just change sort order when sort column id is same.
                                                    then 
                                                        -- Just change sort order.
                                                        case curSortType of
                                                          SortAscending  -> SortDescending
                                                          SortDescending -> SortAscending
                                                    -- Otherwise sort ascending.
                                                    else SortAscending

                     -- Get new sort type.
                     (newSortType, _, _) <- treeSortableGetSortColumnId model

                     -- Update sort status of buffer.
                     modifyIORef (diredViewBuffer view) (\x -> x {diredBufferSortStatus = (option, newSortType)})

                     -- Focus to cell.
                     treeViewFocus (diredViewTreeView view)
                ) (lookup option options) 

-- | Internal sort function.
diredViewSortInternal :: DiredView -> FileInfoOption -> SortType -> IO ()
diredViewSortInternal view option sortType = do
  options <- (<<<=) diredBufferFileInfoOptions diredViewGetBuffer view 
  applyMaybeTM_ (\id -> treeSortableSetSortColumnId (diredViewSortModel view) id sortType) (lookup option options)

-- | Get name.
diredViewGetName :: DiredView -> IO String
diredViewGetName = readIORef . diredViewName

-- | Set name.
diredViewSetName :: DiredView -> String -> IO ()
diredViewSetName = 
  writeIORef . diredViewName

-- | Get buffer.
diredViewGetBuffer :: DiredView -> IO DiredBuffer
diredViewGetBuffer = readIORef . diredViewBuffer

-- | Set buffer.
diredViewSetBuffer :: DiredView -> DiredBuffer -> IO ()
diredViewSetBuffer view buffer = do
  writeIORef (diredViewBuffer view) buffer 
  diredViewDraw view

-- | Keymap.
diredViewKeymap :: Keymap
diredViewKeymap = M.fromList $ map (second (Action . DiredViewAction))
                    [("j",      diredViewNextNode)
                    ,("k",      diredViewPrevNode)
                    ,("f",      diredViewEntryNode)
                    ,("'",      diredViewUpperDirectory)
                    ,("z",      diredViewSortByName)
                    ,("s",      diredViewSortBySize)
                    ,("d",      diredViewSortByItem)
                    ,("x",      diredViewSortByMime)
                    ,("m",      diredViewSortByMTime)
                    ,("a",      diredViewSortByATime)
                    ,("c",      diredViewSortByCTime)]

-- | Apply focus.
diredViewApplyAction :: (DiredView -> IORefObject -> IO ()) -> IORefObject -> IO ()
diredViewApplyAction f ioRefObject =
    applyMaybeTM_ (\Page {pageView = x} -> 
                       applyMaybeTM_ (\d -> f d ioRefObject) (cast x :: Maybe DiredView)) 
        =<< ioRefObjectGetFocusPage ioRefObject

-- | Scroll line up.
diredViewScrollLineUp :: DiredView -> IO ()
diredViewScrollLineUp =
    treeViewScrollLineUp . diredViewTreeView

-- | Scroll line down.
diredViewScrollLineDown :: DiredView -> IO ()
diredViewScrollLineDown =
    treeViewScrollLineDown . diredViewTreeView

-- | Scroll page up.
diredViewScrollPageUp :: DiredView -> IO ()
diredViewScrollPageUp =
    treeViewScrollPageUp . diredViewTreeView

-- | Scroll page down.
diredViewScrollPageDown :: DiredView -> IO ()
diredViewScrollPageDown =
    treeViewScrollPageDown . diredViewTreeView

-- | Begin.
diredViewBegin :: DiredView -> IO ()
diredViewBegin = 
    treeViewBegin . diredViewTreeView

-- | End.
diredViewEnd :: DiredView -> IO ()
diredViewEnd = 
    treeViewEnd . diredViewTreeView 

data DiredViewState =
    DiredViewState {diredViewStateFilename :: String}
                   deriving Typeable

instance Binary DiredViewState where
    put (DiredViewState s) = do
      put (typeIdOf DiredViewState)
      put s
    get = do
      s <- get
      return (DiredViewState s)

instance PageViewState DiredViewState where
    rebuild s = 
        diredViewRebuildState (diredViewStateFilename s) . Just

-- | Rebuild state.
diredViewRebuildState :: FilePath -> Maybe Int -> IORefObject -> IO ()
diredViewRebuildState path id ioRefObject = 
    applyMaybeTM_ (\view -> do
                     -- Create dired view.
                     diredView <- diredViewNew path
                     -- Create new page.
                     page <- pageAdd view id path diredView ioRefObject
                     -- Focus to current page.
                     tabGroupFocusPage (viewTabGroup view) page
                  ) =<< ioRefObjectGetFocusView ioRefObject
  
