{-# LANGUAGE ExistentialQuantification, RankNTypes, NoMonomorphismRestriction, GADTs #-}
module Manatee.PageView.PageView where

import Data.Typeable
import Data.Binary
import Data.Map (Map)

import Graphics.UI.Gtk hiding (Statusbar, statusbarNew, get)

import {-# source #-} Manatee.PageView.DiredView
import {-# source #-} Manatee.PageView.StringView
import {-# source #-} Manatee.PageView.TerminalView
import {-# source #-} Manatee.Keymap.Keymap
import Manatee.IORefObject
import Manatee.Data.TypeId
import Manatee.Widget.Statusbar
import Manatee.UI.Interactivebar
import Manatee.Widget.Outputbar
import Manatee.Utils.Generic

import qualified Data.Map as M

class Typeable a => PageView a where
    pageViewGetName             :: a -> IO String               -- get name
    pageViewSetName             :: a -> String -> IO ()          -- set name
    pageViewUIFrame             :: a -> PageViewUIFrame         -- UI frame.
    pageViewGetKeymap           :: a -> Keymap                  -- get keymap
    pageViewClone               :: a -> IO a                    -- clone
    pageViewFocus               :: a -> IO ()                   -- focus
    pageViewScrollLineUp        :: a -> IO ()                   -- scroll line up
    pageViewScrollLineDown      :: a -> IO ()                   -- scroll line down
    pageViewScrollPageUp        :: a -> IO ()                   -- scroll page up
    pageViewScrollPageDown      :: a -> IO ()                   -- scroll page down
    pageViewBegin               :: a -> IO ()                   -- move to begin
    pageViewEnd                 :: a -> IO ()                   -- move to end
    pageViewGetState            :: a -> IO PageViewStateWrap    -- get state for reboot
    
class (Typeable s, Binary s) => PageViewState s where 
    rebuild     :: s -> Int -> IORefObject -> IO ()

data PageViewStateWrap = forall s . PageViewState s =>
    PageViewStateWrap s

data PageViewStateDict a where
   Dict :: PageViewState s => PageViewStateDict s

data Exists f = forall a . Exists (f a)

-- | Wrap type for 'PageViewState'.
typeOfDict :: PageViewStateDict s -> s
typeOfDict = undefined

-- | Utils map for lookup `PageViewState'.
pageViewStateMap :: Map TypeId (Exists PageViewStateDict)
pageViewStateMap = 
    M.fromList 
         [(typeIdOf StringViewState, Exists (Dict :: PageViewStateDict StringViewState))
         ,(typeIdOf TerminalViewState, Exists (Dict :: PageViewStateDict TerminalViewState))
         ,(typeIdOf DiredViewState,  Exists (Dict :: PageViewStateDict DiredViewState))]

instance Binary PageViewStateWrap where
    put (PageViewStateWrap s) = put s
    get = do
      t <- get
      case M.lookup t pageViewStateMap of
        Nothing -> fail $ "No entry matches the tag " ++ show t
        Just (Exists d@Dict) -> do
            e <- get
            return (PageViewStateWrap (e `asTypeOf` typeOfDict d))

data PageViewUIFrame =
    PageViewUIFrame {pageViewUIFrameBox                 :: VBox      -- box for contain `PageView'
                    ,pageViewUIFrameOutputbar           :: Outputbar -- outpubar for display message
                    ,pageViewUIFrameInteractivebar      :: Interactivebar -- interactivebar for interactive input
                    ,pageViewUIFrameScrolledWindow      :: ScrolledWindow               -- scrolledWindow 
                    ,pageViewUIFrameStatusbar           :: Statusbar                    -- statusbar
                    }

-- | Create 'PageViewUIFrame'.
pageViewUIFrameNew :: Maybe Statusbar -> IO PageViewUIFrame
pageViewUIFrameNew ms = do
  -- Create box.
  box <- vBoxNew False 0

  -- Outputbar.
  outputbar <- outputbarNew

  -- Interactivebar.
  bar <- interactivebarNew

  -- Scroll window.
  scrolledWindow <- scrolledWindowNew Nothing Nothing
  scrolledWindowSetPolicy scrolledWindow PolicyAutomatic PolicyAutomatic 
  boxPackStart box scrolledWindow PackGrow 0 
  
  -- Statusbar.
  statusbar <- statusbarNew box ms 

  return $ PageViewUIFrame box outputbar bar scrolledWindow statusbar
  
-- | Get box for contain 'PageView'.
pageViewBox :: forall a . PageView a => a -> VBox 
pageViewBox = pageViewUIFrameBox . pageViewUIFrame

-- | Get scrolledwindow for 'PageView'
pageViewScrolledWindow :: forall a . PageView a => a -> ScrolledWindow
pageViewScrolledWindow = pageViewUIFrameScrolledWindow . pageViewUIFrame

-- | Update output.
pageViewUpdateOutputStatus :: forall a . PageView a => a -> String -> IO ()
pageViewUpdateOutputStatus pv str = do
  -- Init.
  let frame = pageViewUIFrame pv
      box   = pageViewUIFrameBox frame
      bar   = pageViewUIFrameOutputbar frame
      
  -- Update message.
  outputbarShow box bar str

-- | Update status.
pageViewUpdateInfoStatus :: forall a . PageView a => a -> String -> String -> IO ()
pageViewUpdateInfoStatus = statusbarInfoItemUpdate . pageViewUIFrameStatusbar . pageViewUIFrame

-- | Whether focus on interactivebar.
pageViewIsFocusInteractivebar :: PageViewUIFrame -> IO Bool
pageViewIsFocusInteractivebar = 
    widgetGetIsFocus . interactivebarEntry . pageViewUIFrameInteractivebar 

-- | Whether outputbar is show.
pageViewIsOutputbarShow :: PageViewUIFrame -> IO Bool
pageViewIsOutputbarShow frame = do
  parent <- widgetGetParent $ outputbarFrame $ pageViewUIFrameOutputbar frame
  return $ applyMaybeB parent True

-- | Switch focus between page and interactivar. 
pageViewSwitchFocus :: forall a . PageView a => a -> IO ()
pageViewSwitchFocus pv = do
  let frame = pageViewUIFrame pv
  ifM (pageViewIsFocusInteractivebar frame)
      -- Focus page when interactivebar is active.
      (pageViewFocus pv)
      -- Otherwise try to focus interactivebar.
      (widgetGrabFocus $ interactivebarEntry $ pageViewUIFrameInteractivebar frame)
