{-# LANGUAGE ExistentialQuantification, RankNTypes, DeriveDataTypeable, ScopedTypeVariables #-}
{-# LANGUAGE TemplateHaskell #-}
module Manatee.PageView.StringView where

import Data.DeriveTH
import Control.Monad
import Control.Applicative
import Text.Printf
import Data.Typeable
import Control.Arrow
import Data.Binary

import Graphics.UI.Gtk hiding (Statusbar, statusbarNew, get)
import Graphics.UI.Gtk.SourceView.SourceView
import Graphics.UI.Gtk.SourceView.SourceBuffer
import Graphics.UI.Gtk.SourceView.SourceLanguage
import Graphics.UI.Gtk.SourceView.SourceLanguageManager

import Manatee.Widget.Statusbar
import Manatee.Utils.Generic
import Manatee.PageView.PageView
import Manatee.UI.Page
import Manatee.UI.View
import Manatee.UI.TabGroup
import Manatee.Utils.Gtk.Gtk
import Manatee.Utils.Gtk.Multiline
import Manatee.IORefObject
import                          {-# source #-} Manatee.Keymap.Keymap
import Manatee.Data.TypeId

import qualified Data.ByteString.Lazy as B
import qualified Data.ByteString.Lazy.UTF8 as U
import qualified Data.Map as M

data StringView =
    StringView {stringViewName            :: String
               ,stringViewUIFrame         :: PageViewUIFrame
               ,stringViewView            :: SourceView
               }
    deriving Typeable

instance Show StringView where
    show = printf "<#StringView %s>" . show . stringViewName

instance PageView StringView where
    pageViewGetName = return . stringViewName

    pageViewSetName _ _ = return ()

    pageViewUIFrame = stringViewUIFrame

    pageViewClone = stringViewClone

    pageViewFocus = widgetGrabFocus . stringViewView

    pageViewScrollLineUp = stringViewScrollLineUp
    
    pageViewScrollLineDown = stringViewScrollLineDown
    
    pageViewScrollPageUp = stringViewScrollPageUp
    
    pageViewScrollPageDown = stringViewScrollPageDown

    pageViewBegin = stringViewBegin
    
    pageViewEnd = stringViewEnd

    pageViewGetKeymap _ = stringViewKeymap

    pageViewGetState a = do
      let name = stringViewName a
      selectionStatus <- stringViewGetSelectionStatus a
      return $ PageViewStateWrap $ StringViewState name selectionStatus

data StringViewAction = StringViewAction (StringView -> IO ())
data StringViewEditAction = StringViewEditAction (StringView -> IO ())

instance ActionExecute StringViewAction where
    execute (StringViewAction a) = stringViewApplyAction a

instance ActionExecute StringViewEditAction where
    execute (StringViewEditAction a) = stringViewApplyEditAction a

stringViewScrolledWindow :: StringView -> ScrolledWindow
stringViewScrolledWindow = pageViewUIFrameScrolledWindow . stringViewUIFrame

stringViewStatusbar :: StringView -> Statusbar
stringViewStatusbar = pageViewUIFrameStatusbar . stringViewUIFrame

-- | Create string buffer.
stringViewNew :: String -> IO StringView
stringViewNew name = do
  -- Get new string buffer.
  stringView <- stringViewNewInternal name Nothing Nothing

  -- Syntax highglight.
  stringViewSyntaxHighlight stringView

  return stringView

-- | Clone string buffer.
stringViewClone :: StringView -> IO StringView
stringViewClone sb = do
  sourceBuffer <- stringViewGetSourceBuffer sb
  let name = stringViewName sb
  stringViewNewInternal name (Just sourceBuffer) (Just $ stringViewStatusbar sb)

-- | Internal function for create string buffer.
stringViewNewInternal :: String -> Maybe SourceBuffer -> Maybe Statusbar -> IO StringView
stringViewNewInternal name sourceBuffer statusbarClone = do
  -- Create UI frame.
  uiFrame <- pageViewUIFrameNew statusbarClone
  let scrolledWindow = pageViewUIFrameScrolledWindow uiFrame

  sourceView <- stringViewViewNew sourceBuffer
  scrolledWindow `containerAdd` sourceView
  let stringView = StringView name uiFrame sourceView

  -- Customize setup.
  sourceViewSetHighlightCurrentLine sourceView True      -- highglight current line.
  sourceViewSetInsertSpacesInsteadOfTabs sourceView True -- use space instead tabs
  sourceViewSetShowLineNumbers sourceView True

  -- Update position status.
  stringViewHandlePositionDisplay stringView -- display position with user action

  -- Handle selection mark.
  stringViewHandleSelectionMark stringView

  -- Handle edit track.
  stringViewHandleEditTrack stringView

  return stringView

-- | Create string buffer view
stringViewViewNew :: Maybe SourceBuffer -> IO SourceView
stringViewViewNew sb = 
    -- Source view.
    case sb of
      Just x -> do
              sourceBufferSetHighlightMatchingBrackets x True -- highglight match brackets
              sourceBufferSetMaxUndoLevels x (-1)             -- no limint for undo level
              sourceViewNewWithBuffer x                       -- create soruce buffer by user 
      Nothing -> sourceViewNew                                -- create new source buffer

-- | Get text buffer.
stringViewGetTextBuffer :: StringView -> IO TextBuffer 
stringViewGetTextBuffer = textViewGetBuffer . stringViewView

-- | Get source buffer.
stringViewGetSourceBuffer :: StringView -> IO SourceBuffer
stringViewGetSourceBuffer sb = 
    castToSourceBuffer <$> stringViewGetTextBuffer sb

-- | Get buffer content.
stringViewGetText :: StringView -> IO String
stringViewGetText = textViewGetText . stringViewView

-- | Get current line.
stringViewGetLine :: StringView -> IO Int
stringViewGetLine = textViewGetLine . stringViewView

-- | Get current column.
stringViewGetColumn :: StringView -> IO Int
stringViewGetColumn = textViewGetColumn . stringViewView

-- | Display position information.
stringViewDisplayPositionStatus :: StringView -> IO ()
stringViewDisplayPositionStatus sb = 
  liftM2 (printf "(%d, %d)") (stringViewGetLine sb) (stringViewGetColumn sb)
             >>= statusbarInfoItemUpdate (stringViewStatusbar sb) "Position" 
  >> return ()

-- | Update position information.
stringViewHandlePositionDisplay :: StringView -> IO ()
stringViewHandlePositionDisplay sv = do
  stringViewView sv `afterExposeRect` (\_ -> stringViewDisplayPositionStatus sv)
  return ()

-- | Handle selection mark.
stringViewHandleSelectionMark :: StringView -> IO ()  
stringViewHandleSelectionMark sv = do
  -- Get source view.
  let sourceView = stringViewView sv 

  -- Cancel selection when button press.
  onButtonPress sourceView (\_ -> textViewCancelSelectionMark sourceView >> return False)

  -- Update selection iter after buffer changed.
  buffer <- textViewGetBuffer sourceView
  onBufferChanged buffer (textBufferUpdateSelectionIter buffer)

  -- Apply selection iter when focus in.
  onFocusIn sourceView (\_ -> stringViewApplySelectionMark sv >> return False)

  return ()

-- | Handle edit track.
stringViewHandleEditTrack :: StringView -> IO ()  
stringViewHandleEditTrack sv = do
  -- Get source view and buffer.
  let sourceView = stringViewView sv
  buffer <- textViewGetBuffer sourceView
  
  -- Add edit mark after buffer changed.
  afterBufferChanged buffer $ textBufferAddEditTrack buffer

  return ()

-- | Get string buffer language.
stringViewGetLanguage :: StringView -> IO SourceLanguage
stringViewGetLanguage buffer = 
    sourceBufferGetLanguage
        =<< stringViewGetSourceBuffer buffer

-- | Set string buffer language.
stringViewSetLanguage :: StringView -> SourceLanguage -> IO ()
stringViewSetLanguage buffer language = 
    (<=<) (\sb -> sourceBufferSetLanguage sb language) stringViewGetSourceBuffer buffer

-- | Syntax highlight with file name.
stringViewSyntaxHighlight :: StringView -> IO ()
stringViewSyntaxHighlight buffer = do
  lm <- sourceLanguageManagerNew
  (_, lang) <- sourceLanguageForFilename lm (Just $ stringViewName buffer)
  applyMaybeTM_ (stringViewSetLanguage buffer) lang

-- | Save buffer.
stringViewSave :: StringView -> IO ()
stringViewSave a = do
  byteString <- (<<<=) U.fromString stringViewGetText a
  let filePath = stringViewName a
  B.writeFile filePath byteString
  pageViewUpdateOutputStatus a ("Save " ++ stringViewName a)

-- | Undo.
stringViewUndo :: StringView -> IO ()
stringViewUndo a = do
  sb <- stringViewGetSourceBuffer a
  ifM (sourceBufferGetCanUndo sb)
          (do
            sourceBufferUndo sb
            pageViewUpdateOutputStatus a "Undo!")
          (pageViewUpdateOutputStatus a "No futher undo information.")

-- | Redo.
stringViewRedo :: StringView -> IO ()
stringViewRedo a = do
  sb <- stringViewGetSourceBuffer a
  ifM (sourceBufferGetCanRedo sb)
          (do
            sourceBufferRedo sb
            pageViewUpdateOutputStatus a "Redo!")
          (pageViewUpdateOutputStatus a "No futher redo information.")

-- | String view wrap user action for undo/redo.
stringViewWrapAction :: StringView -> IO () -> IO ()  
stringViewWrapAction  = textViewWrapAction . stringViewView

-- | Newline.
stringViewNewline :: StringView -> IO ()  
stringViewNewline = textViewNewLine . stringViewView

-- | Open newline below.
stringViewOpenNewlineBelow :: StringView -> IO ()
stringViewOpenNewlineBelow a = 
    textViewOpenNewlineBelow (stringViewView a) (stringViewScrolledWindow a)

-- | Open newline above.
stringViewOpenNewlineAbove :: StringView -> IO ()
stringViewOpenNewlineAbove a = 
    textViewOpenNewlineAbove (stringViewView a) (stringViewScrolledWindow a)

-- | Select all.
stringViewSelectAll :: StringView -> IO ()
stringViewSelectAll = textViewSelectAll . stringViewView

-- | Delete.
stringViewDelete :: StringView -> IO ()
stringViewDelete = textViewDelete . stringViewView

-- | Cut.
stringViewCut :: StringView -> IO ()
stringViewCut = textViewCut . stringViewView

-- | Copy.
stringViewCopy :: StringView -> IO ()
stringViewCopy = textViewCopy . stringViewView

-- | Paste.
stringViewPaste :: StringView -> IO ()
stringViewPaste = textViewPaste . stringViewView

-- | Forward line.
stringViewForwardLine :: StringView -> IO ()
stringViewForwardLine a = do 
  textViewForwardLine (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Backward line.
stringViewBackwardLine :: StringView -> IO ()
stringViewBackwardLine a = do 
  textViewBackwardLine (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Forward char.
stringViewForwardChar :: StringView -> IO ()
stringViewForwardChar a = do 
  textViewForwardChar (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Backward char.
stringViewBackwardChar :: StringView -> IO ()
stringViewBackwardChar a = do 
  textViewBackwardChar (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Forward word.
stringViewForwardWord :: StringView -> IO ()
stringViewForwardWord a = do 
  textViewForwardWord (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Backward word.
stringViewBackwardWord :: StringView -> IO ()
stringViewBackwardWord a = do 
  textViewBackwardWord (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Begin.
stringViewBegin :: StringView -> IO ()
stringViewBegin a = do 
  textViewBegin (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | End.
stringViewEnd :: StringView -> IO ()
stringViewEnd a = do 
  textViewEnd (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Scroll line up.
stringViewScrollLineUp :: StringView -> IO ()
stringViewScrollLineUp a = do 
  textViewScrollLineUp (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Scroll line down.
stringViewScrollLineDown :: StringView -> IO ()
stringViewScrollLineDown a = do 
  textViewScrollLineDown (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Scroll page up.
stringViewScrollPageUp :: StringView -> IO ()
stringViewScrollPageUp a = do 
  textViewScrollPageUp (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Scroll page down.
stringViewScrollPageDown :: StringView -> IO ()
stringViewScrollPageDown a = do 
  textViewScrollPageDown (stringViewView a) (stringViewScrolledWindow a)
  stringViewApplySelectionMark a

-- | Smart home.
stringViewSmartHome :: StringView -> IO ()
stringViewSmartHome a = do 
  textViewSmartHome $ stringViewView a
  stringViewApplySelectionMark a

-- | Smart end.
stringViewSmartEnd :: StringView -> IO ()
stringViewSmartEnd a = do 
  textViewSmartEnd $ stringViewView a
  stringViewApplySelectionMark a

-- | Delete forward char.
stringViewDeleteForwardChar :: StringView -> IO ()
stringViewDeleteForwardChar = textViewDeleteForwardChar . stringViewView

-- | Backward char.
stringViewDeleteBackwardChar :: StringView -> IO ()
stringViewDeleteBackwardChar = textViewDeleteBackwardChar . stringViewView

-- | Forward word.
stringViewDeleteForwardWord :: StringView -> IO ()
stringViewDeleteForwardWord = textViewDeleteForwardWord . stringViewView

-- | Backward word.
stringViewDeleteBackwardWord :: StringView -> IO ()
stringViewDeleteBackwardWord = textViewDeleteBackwardWord . stringViewView

-- | Delete to line end.
stringViewDeleteToLineEnd :: StringView -> IO ()
stringViewDeleteToLineEnd = textViewDeleteToLineEnd . stringViewView

-- | Delete to line start.
stringViewDeleteToLineStart :: StringView -> IO ()
stringViewDeleteToLineStart = textViewDeleteToLineStart . stringViewView

-- | Duplicate lines.
stringViewDupLinesBelow, stringViewDupLinesAbove :: StringView -> IO ()
stringViewDupLinesBelow = textViewDupLinesBelow . stringViewView
stringViewDupLinesAbove = textViewDupLinesAbove . stringViewView

-- | Delete lines.
stringViewDelLines :: StringView -> IO ()
stringViewDelLines = textViewDelLines . stringViewView

-- | Transposes lines.
stringViewTraLinesBelow, stringViewTraLinesAbove :: StringView -> IO ()
stringViewTraLinesBelow = textViewTraLinesBelow . stringViewView
stringViewTraLinesAbove = textViewTraLinesAbove . stringViewView

-- | Move to first edit position.
stringViewMoveToFirstEdit :: StringView -> IO () 
stringViewMoveToFirstEdit a = 
    textViewMoveToFirstEdit (stringViewView a) >> return ()

-- | Move to last edit position.
stringViewMoveToLastEdit :: StringView -> IO () 
stringViewMoveToLastEdit a = 
    textViewMoveToLastEdit (stringViewView a) >> return ()

-- | Move to previous edit postion.
stringViewMoveToPrevEdit :: StringView -> IO ()
stringViewMoveToPrevEdit a = 
    textViewMoveToPrevEdit (stringViewView a) >> return ()

-- | Move to next edit postion.
stringViewMoveToNextEdit :: StringView -> IO ()
stringViewMoveToNextEdit a = 
    textViewMoveToNextEdit (stringViewView a) >> return ()

-- | Reload file.
stringViewReload :: StringView -> IO ()
stringViewReload sv = do
  -- Reload file.
  textViewLoadFile (stringViewView sv) (stringViewName sv)

  -- Move to view begin.
  stringViewBegin sv

-- | Set text.
stringViewSetText :: StringView -> String -> IO ()
stringViewSetText = textViewSetText . stringViewView

-- | String buffer keymap.
stringViewKeymap :: Keymap
stringViewKeymap = 
    M.fromList 
         $ map (second (Action . StringViewAction))
         [("M-a",       stringViewSelectAll)
         ,("M-s",       stringViewSave)
         ,("M-x",       stringViewCopy)
         ,("M-j",       stringViewForwardLine)
         ,("M-k",       stringViewBackwardLine)
         ,("Down",      stringViewForwardLine)
         ,("Up",        stringViewBackwardLine)
         ,("M-l",       stringViewForwardChar)
         ,("M-h",       stringViewBackwardChar)
         ,("M-L",       stringViewForwardWord)
         ,("M-H",       stringViewBackwardWord)
         ,("M-y",       stringViewSmartHome)
         ,("M-o",       stringViewSmartEnd)
         ,("M-/",       stringViewUndo)
         ,("M-?",       stringViewRedo)
         ,("M-r",       stringViewReload)
         ,("M-c",       stringViewToggleSelectionMark)
         ,("M-C",       stringViewExchangeSelectionMark)
         ,("C-n",       stringViewMoveToNextEdit)
         ,("C-m",       stringViewMoveToPrevEdit)
         ,("C-N",       stringViewMoveToLastEdit)
         ,("C-M",       stringViewMoveToFirstEdit)]
         ++ map (second (Action . StringViewEditAction))
         [("RET",       stringViewNewline)
         ,("M-n",       stringViewOpenNewlineBelow)
         ,("M-m",       stringViewOpenNewlineAbove)
         ,("M-z",       stringViewDelete)
         ,("M-Z",       stringViewCut)
         ,("M-X",       stringViewPaste)
         ,("M-d",       stringViewDupLinesBelow)
         ,("M-D",       stringViewDupLinesAbove)
         ,("M-f",       stringViewDelLines)
         ,("M-e",       stringViewTraLinesBelow)
         ,("M-E",       stringViewTraLinesAbove)
         ,("M-.",       stringViewDeleteForwardChar)
         ,("M-,",       stringViewDeleteBackwardChar)
         ,("M->",       stringViewDeleteForwardWord)
         ,("M-<",       stringViewDeleteBackwardWord)
         ,("M-M",       stringViewDeleteToLineEnd)
         ,("M-N",       stringViewDeleteToLineStart)]

-- | Apply action.
stringViewApplyAction :: (StringView -> IO ()) -> IORefObject -> IO ()
stringViewApplyAction f ioRefObject =
    applyMaybeTM_ (\Page {pageView = x} -> 
                   -- Apply action.
                       applyMaybeTM_ f (cast x :: Maybe StringView)) 
        =<< ioRefObjectGetFocusPage ioRefObject

-- | Apply edit action.
-- Edit action modified buffer, so edit action can undo/redo.
stringViewApplyEditAction :: (StringView -> IO ()) -> IORefObject -> IO ()
stringViewApplyEditAction f ioRefObject =
    applyMaybeTM_ (\Page {pageView = x} -> 
                   -- Apply edit action.
                       applyMaybeTM_ (\sv -> stringViewWrapAction sv $ f sv
                                     ) (cast x :: Maybe StringView)) 
        =<< ioRefObjectGetFocusPage ioRefObject

-- | Set selection mark.
stringViewToggleSelectionMark :: StringView -> IO ()
stringViewToggleSelectionMark = textViewToggleSelectionMark . stringViewView 

-- | Exchange selection mark.
stringViewExchangeSelectionMark :: StringView -> IO ()
stringViewExchangeSelectionMark = textViewExchangeSelectionMark . stringViewView 

-- | Show selection mark.
stringViewApplySelectionMark :: StringView -> IO ()
stringViewApplySelectionMark = textViewApplySelectionMark . stringViewView

-- | Get selection status.
stringViewGetSelectionStatus :: StringView -> IO (Maybe SelectionStatus) 
stringViewGetSelectionStatus sv = do
  -- Get text buffer.
  tb <- (<<=<) textViewGetBuffer stringViewView sv

  -- Save selection status when selectin iter is exist.
  -- Don't use `textBufferHasSelection`, if you split buffer
  -- with many view, it can't work when other non-focus view 
  -- no `redraw` selection area.
  selectionIter <- textBufferGetSelectionIter tb 
  applyMaybeTM (\i -> do
                  -- Get insert iter.
                  iter <- textBufferGetTextIter tb
                  -- Transform iter to offset.
                  insOffset <- textIterGetOffset iter
                  bouOffset <- textIterGetOffset i
                  -- Get selecton status.
                  return $ Just $ SelectionStatus insOffset bouOffset
               ) selectionIter

data SelectionStatus = 
    SelectionStatus {ssInsertOffset    :: Int
                    ,ssBoundOffset     :: Int}
                     deriving Typeable

data StringViewState =
    StringViewState {stringViewStateFilename    :: String
                    ,stringViewStateSelection   :: Maybe SelectionStatus} 
                    deriving Typeable

instance Binary StringViewState where
    put (StringViewState n s) = do
      put (typeIdOf StringViewState)
      put n
      put s
    get = do
      n <- get
      s <- get
      return (StringViewState n s)

instance PageViewState StringViewState where
    rebuild s = 
        stringViewRebuildState s . Just

-- | Rebuild state.
stringViewRebuildState :: StringViewState -> Maybe Int -> IORefObject -> IO ()
stringViewRebuildState svs id ioRefObject = 
    applyMaybeTM_ (\view -> do
                     -- Get status.
                     let path = stringViewStateFilename svs 
                         selection = stringViewStateSelection svs
                     -- Get string view.
                     stringView <- stringViewNew path
                     -- Load file content.
                     textViewLoadFile (stringViewView stringView) path
                     -- Create page.
                     page <- pageAdd view id path stringView ioRefObject
                     -- Move to view begin.
                     stringViewBegin stringView
                     -- Restore selection.
                     applyMaybeTM_ (\(SelectionStatus insOffset bouOffset) -> do
                                      -- Get text buffer.
                                      tb <- (<<=<) textViewGetBuffer stringViewView stringView
                                      -- Restore iter from offset.
                                      insIter <- textBufferGetIterAtOffset tb insOffset
                                      bouIter <- textBufferGetIterAtOffset tb bouOffset
                                      -- Sleection range.
                                      textBufferSelectRange tb insIter bouIter
                                      -- Update selection iter.
                                      textBufferUpdateSelectionIter tb
                                   ) selection
                     -- Focus to current page.
                     tabGroupFocusPage (viewTabGroup view) page
                  ) =<< ioRefObjectGetFocusView ioRefObject
  
$(derive makeBinary ''SelectionStatus)
