{-# LANGUAGE ExistentialQuantification, RankNTypes, DeriveDataTypeable, ScopedTypeVariables #-}
module Manatee.PageView.TerminalView where

import Control.Monad
import Data.IORef
import Text.Printf
import Data.Typeable
import Data.Binary

import Graphics.UI.Gtk hiding (Statusbar, statusbarNew, get)
import Graphics.UI.Gtk.Scrolling.ScrolledWindow
import Graphics.UI.Gtk.Vte.Vte

import Manatee.Utils.Generic
import Manatee.Utils.Gtk.Gtk
import Manatee.PageView.PageView
import Manatee.UI.Page
import Manatee.UI.View
import Manatee.UI.TabGroup
import Manatee.IORefObject
import {-# source #-} Manatee.Keymap.Keymap
import Manatee.Data.TypeId

import qualified Data.Map as M

data TerminalView =
    TerminalView {terminalViewName              :: String
                 ,terminalViewUIFrame           :: PageViewUIFrame
                 ,terminalViewTerminal          :: Terminal
                 }
    deriving Typeable

instance Show TerminalView where
    show = printf "<#TerminalView %s>" . show . terminalViewName
    
instance PageView TerminalView where
    pageViewGetName = return . terminalViewName

    pageViewSetName _ _ = return ()

    pageViewUIFrame = terminalViewUIFrame

    pageViewClone = terminalViewClone

    pageViewFocus = widgetGrabFocus . terminalViewTerminal

    pageViewScrollLineUp = terminalViewScrollLineUp
    
    pageViewScrollLineDown = terminalViewScrollLineDown
    
    pageViewScrollPageUp = terminalViewScrollPageUp
    
    pageViewScrollPageDown = terminalViewScrollPageDown

    pageViewBegin = terminalViewBegin
    
    pageViewEnd = terminalViewEnd

    pageViewGetKeymap _ = terminalViewKeymap

    pageViewGetState = return . PageViewStateWrap . TerminalViewState . terminalViewName
    
data TerminalViewAction = TerminalViewAction (TerminalView -> IO ())

instance ActionExecute TerminalViewAction where
    execute (TerminalViewAction a) = terminalViewApplyAction a

-- | Create terminal view.
terminalViewNew :: String -> IO TerminalView    
terminalViewNew name = do
  -- Create UI frame.
  uiFrame <- pageViewUIFrameNew Nothing
  let scrolledWindow = pageViewUIFrameScrolledWindow uiFrame
      box = pageViewUIFrameBox uiFrame
      
  -- Terminal.
  terminal <- terminalNew
  terminalForkCommand terminal Nothing Nothing Nothing Nothing False False False
  font <- fontDescriptionFromString "DejaVu Sans YuanTi Mono 10"
  terminalSetFont terminal font
  terminalSetBackgroundImageFile terminal "/data/Picture/壁纸/vladstudio_venice_grandcanal_1280x1024.jpg"
  scrolledWindow `containerAdd` terminal
  on terminal childExited $ widgetDestroy box -- destroy box when exit
                                              -- page will be remove when box destory

  -- Terminal view.
  let terminalView = TerminalView name uiFrame terminal
      
  return terminalView

terminalViewScrolledWindow :: TerminalView -> ScrolledWindow
terminalViewScrolledWindow = pageViewUIFrameScrolledWindow . terminalViewUIFrame
  
-- | Clone terminal view.
terminalViewClone :: TerminalView -> IO TerminalView  
terminalViewClone = terminalViewNew . terminalViewName

-- | Temrinal view keymap.
terminalViewKeymap :: Keymap
terminalViewKeymap = 
    M.fromList []

-- | Apply focus.
terminalViewApplyAction :: (TerminalView -> IO ()) -> IORefObject -> IO ()
terminalViewApplyAction f ioRefObject =
    applyMaybeTM_ (\Page {pageView = x} -> 
                       applyMaybeTM_ f (cast x :: Maybe TerminalView)) 
        =<< ioRefObjectGetFocusPage ioRefObject

-- | Scroll line up.
terminalViewScrollLineUp :: TerminalView -> IO ()
terminalViewScrollLineUp tv = terminalScrollLineUp (terminalViewTerminal tv) (terminalViewScrolledWindow tv)

-- | Scroll line down.
terminalViewScrollLineDown :: TerminalView -> IO ()
terminalViewScrollLineDown tv = terminalScrollLineDown (terminalViewTerminal tv) (terminalViewScrolledWindow tv)

-- | Scroll page up.
terminalViewScrollPageUp :: TerminalView -> IO ()
terminalViewScrollPageUp tv = terminalScrollPageUp (terminalViewTerminal tv) (terminalViewScrolledWindow tv)

-- | Scroll page down.
terminalViewScrollPageDown :: TerminalView -> IO ()
terminalViewScrollPageDown tv = terminalScrollPageDown (terminalViewTerminal tv) (terminalViewScrolledWindow tv)

-- | Move begin.
terminalViewBegin :: TerminalView -> IO ()
terminalViewBegin = terminalBegin . terminalViewScrolledWindow

-- | Move end.
terminalViewEnd :: TerminalView -> IO ()
terminalViewEnd = terminalEnd . terminalViewScrolledWindow

-- | Generate unique name.
terminalGenerateName :: IORefObject -> IO String
terminalGenerateName ioRefObject = do
  let terminalCounter = ioRefObjectTerminalCounter ioRefObject
  modifyIORef terminalCounter succ 
  counter <- readIORef terminalCounter
  let name = "<*Terminal" ++ show counter ++ "*>" 
  return name

data TerminalViewState =
    TerminalViewState {terminalViewStateName    :: String} 
                    deriving Typeable

instance Binary TerminalViewState where
    put (TerminalViewState s) = do
      put (typeIdOf TerminalViewState)
      put s
    get = do
      s <- get
      return (TerminalViewState s)

instance PageViewState TerminalViewState where
    rebuild s = 
        terminalViewRebuildState (terminalViewStateName s) . Just

-- | Rebuild state.
terminalViewRebuildState :: String -> Maybe Int -> IORefObject -> IO ()
terminalViewRebuildState name id ioRefObject = 
    applyMaybeTM_ (\view -> do
                     -- Get terminal view.
                     terminalView <- terminalViewNew name
                     -- Create page.
                     page <- pageAdd view id name terminalView ioRefObject
                     -- Focus to current page.
                     tabGroupFocusPage (viewTabGroup view) page
                  ) =<< ioRefObjectGetFocusView ioRefObject
  
  
