{-# LANGUAGE DeriveDataTypeable, TemplateHaskell #-}
module Manatee.State.ViewNode where

import Data.DeriveTH
import Data.Binary
import Data.Function
import Data.Set (Set)
import Data.IORef
import Control.Monad
import Data.Ord

import Manatee.Data.SetList
import Manatee.Utils.Set
import Manatee.Utils.Generic
import Manatee.UI.ViewNode
import Manatee.IORefObject
import Manatee.UI.View

import qualified Manatee.Utils.Set as MUS 

data ViewNodeState =
    ViewNodeState {vnsId                :: Int
                  ,vnsParentId          :: Maybe Int
                  ,vnsChildLeftId       :: Maybe Int
                  ,vnsChildRightId      :: Maybe Int
                  ,vnsType              :: ViewNodeType
                  ,vnsDirection         :: ViewNodeDirection}

data ViewNodeListState =
    ViewNodeListState {vnlsSet           :: Set ViewNodeState
                      ,vnlsCounter       :: Int}

instance Eq ViewNodeState where
    (==) = (==) `on` vnsId

instance Ord ViewNodeState where
    compare = comparing vnsId

-- | Get view node state.
viewNodeStateGet :: ViewNode -> IO ViewNodeState
viewNodeStateGet vn = do
  let id   = viewNodeId vn
      dire = viewNodeDirection vn
  pId  <- readIORef $ viewNodeParentId vn
  clId <- readIORef $ viewNodeChildLeftId vn
  crId <- readIORef $ viewNodeChildRightId vn
  typ  <- readIORef $ viewNodeType vn
  return $ ViewNodeState id pId clId crId typ dire

-- | Get view node list state.
viewNodeListStateGet :: IORefObject -> IO ViewNodeListState
viewNodeListStateGet ioRefObject = do
  -- Get init value.
  viewNodeList <- readIORef $ ioRefObjectViewNodeList ioRefObject

  -- Get view node counter.
  let viewNodeCounter = listCounter viewNodeList
  
  -- Get node state set.
  viewNodeStateSet <- MUS.mapM viewNodeStateGet $ listSet viewNodeList 

  return $ ViewNodeListState viewNodeStateSet viewNodeCounter

-- | Rebuild view node state.
viewNodeStateRebuild :: ViewNodeState -> Set ViewNodeState -> IORefObject -> IO ()
viewNodeStateRebuild state set ioRefObject = do
  -- Get init value.
  let id   = vnsId state
      pId  = vnsParentId state
      clId = vnsChildLeftId state
      crId = vnsChildRightId state
      typ  = vnsType state
      dire = vnsDirection state

  -- New view node.
  viewNode <- viewNodeNewInternal (Just id) pId clId crId typ dire ioRefObject

  -- Try rebuild left child node.
  applyMaybeTM_ (\n -> viewNodeStateRebuild n set ioRefObject
                ) $ applyMaybeT (\i -> maybeFindMin set (\x -> vnsId x == i)) clId

  -- Try rebuild right child node.
  applyMaybeTM_ (\n -> viewNodeStateRebuild n set ioRefObject
                ) $ applyMaybeT (\i -> maybeFindMin set (\x -> vnsId x == i)) crId
  
  -- Build view when current node haven't any children.
  when (clId == Nothing && crId == Nothing) $ 
       viewNewInternal viewNode Nothing ioRefObject
                        >> return ()
  
-- | Rebuild view node list state.
viewNodeListStateRebuild :: ViewNodeListState -> IORefObject -> IO ()
viewNodeListStateRebuild state ioRefObject = do
  -- Get init value.
  let viewNodeList = ioRefObjectViewNodeList ioRefObject
      set          = vnlsSet state
      rootNode     = maybeFindMin set (\x -> vnsType x == TNodeRoot)
      counter      = vnlsCounter state

  -- Init view node list counter.
  modifyIORef viewNodeList (\x -> setListSetCounter x counter)

  -- Rebuild view node.
  case rootNode of
    Just n  -> viewNodeStateRebuild n set ioRefObject
    Nothing -> return ()

-- | Generate binary instance automatically.
$(derive makeBinary ''ViewNodeState)
$(derive makeBinary ''ViewNodeListState)
$(derive makeBinary ''ViewNodeType)
$(derive makeBinary ''ViewNodeDirection)

