{-# LANGUAGE ExistentialQuantification, RankNTypes #-}
module Manatee.UI.Page where

import Text.Printf
import Data.IORef
import Data.Function
import System.FilePath
import Data.List
import Control.Monad
import Data.Ord

import Graphics.UI.Gtk hiding (on, get)

import Manatee.Utils.Generic
import Manatee.PageView.PageView
import Manatee.IORefObject
import Manatee.Data.Unique
import Manatee.Data.SetList
import {-# source #-} Manatee.UI.View
import {-# source #-} Manatee.PageMode.PageMode

data Page = forall a . PageView a => 
    Page {pageName      :: IORef String
         ,pageId        :: Int
         ,pageView      :: a
         ,pageMode      :: PageMode}

data PageInsertPosition = PStart
                        | PEnd
                        | PLeft
                        | PRight
                        deriving (Eq, Show, Read)

data PageOpenStyle = PForeground
                   | PBackground
                   deriving (Eq, Show, Read)

instance Eq Page where
    (==) = (==) `on` pageId

instance Ord Page where
    compare = comparing pageId

instance Show Page where
    show = printf "<#Page %s>" . show . pageId

-- | Add page.
pageAdd :: PageView a => View -> Maybe Int -> String -> a -> IORefObject -> IO Page
pageAdd view pId filepath buffer ioRefObject = do
  let pageList = ioRefObjectPageList ioRefObject
      viewList = ioRefObjectViewList ioRefObject

  -- Create new page.
  pn <- newIORef filepath
  id <- case pId of
         -- Use speical id.
         Just i  -> return i
         -- Otherwise use new page id.
         Nothing -> pageListGetNewId pageList
  let mode = pageModeLookup filepath 
  page <- pageNew pn id buffer mode

  -- Add page. 
  ioRefObjectAddPage page ioRefObject

  -- Unique page name.
  pageNameUnique ioRefObject

  -- Focus view page.
  -- This step is NECESSERY for focus widget in view.
  viewListFocus viewList view 

  return page

-- | Unique page name.
pageNameUnique :: IORefObject -> IO ()
pageNameUnique ioRefObject = do
  -- Get page list.
  pl  <- pageListGetList $ ioRefObjectPageList ioRefObject

  -- Get page name list.
  pnl <- mapM pageGetName pl

  -- Get page buffer name list.
  pbnl <- mapM pageGetBufferName pl

  let -- Get new name list.
      newNameList = map (uncurry pageNameConcat) $ unique pbnl
      -- Get different name list.
      differentNameList = map snd $ different (pairPred pnl) (pairPred newNameList)

  -- Update new name.
  mapM_ (\i -> ioRefObjectSetPageName (pl !! i) (newNameList !! i) ioRefObject) differentNameList

-- | Concat page name.
pageNameConcat :: FilePath -> FilePath -> FilePath
pageNameConcat fileName diffName 
    | null diffName 
        = fileName
    | otherwise     
        = fileName ++ "<" ++ diffName ++ ">"

-- | Get page name.
pageGetName :: Page -> IO String
pageGetName = readIORef . pageName

-- | Set page name.
pageSetName :: Page -> String -> IO ()
pageSetName page = modifyIORef (pageName page) . const

-- | Page box.
pageBox :: Page -> VBox 
pageBox = pageApplyBuffer pageViewBox

-- | Page clone interface.
pageClone :: Page -> IO Page
pageClone page = 
  pageApplyBuffer (\x -> do
                     let id   = pageId page
                         mode = pageMode page
                     name     <- newIORef =<< pageGetName page
                     pBuffer  <- pageViewClone x
                     pageNew name id pBuffer mode) page 

-- | New page.
pageNew :: forall a . PageView a => IORef String -> Int -> a -> PageMode -> IO Page
pageNew name id buffer mode = do
  -- Create page.
  let page = Page name id buffer mode

  -- Local mode load.
  pageModeLoad mode page

  return page

-- | Applying function with page buffer.
pageApplyBuffer :: (forall a . PageView a => a -> b) -> Page -> b
pageApplyBuffer f (Page {pageView = x}) = f x

-- | Get buffer name.
pageGetBufferName :: Page -> IO String
pageGetBufferName = pageApplyBuffer pageViewGetName

-- | Remove page.
pageRemove :: Page -> IORefObject -> IO ()
pageRemove page ioRefObject =  do
  -- Remove page.
  ioRefObjectRemovePage page ioRefObject

  -- Unique page name.
  pageNameUnique ioRefObject

-- | Apply focus page buffer.
pageApplyFocusBuffer :: (forall a . PageView a => a -> IO ()) -> IORefObject -> IO ()
pageApplyFocusBuffer f ioRefObject = 
    applyMaybeTM_ (pageApplyBuffer f) 
        =<< ioRefObjectGetFocusPage ioRefObject

-- | Apply focus page.
pageApplyFocus :: (Page -> IORefObject -> IO ()) -> IORefObject -> IO ()
pageApplyFocus f ioRefObject =
    applyMaybeTM_ (\p -> f p ioRefObject) 
        =<< ioRefObjectGetFocusPage ioRefObject

type PageList = SetList Page

-- | Create page list.
pageListNew :: IO (IORef PageList)
pageListNew = newIORef setListNew

-- | Get counter of page list.
pageListGetCounter :: IORef PageList -> IO Int
pageListGetCounter =
    (<<<=) setListGetCounter readIORef

-- | Increase page list counter.
pageListIncCounter :: IORef PageList -> IO ()
pageListIncCounter pageList =
    modifyIORef pageList setListIncCounter

-- | Get new page id.
pageListGetNewId :: IORef PageList -> IO Int
pageListGetNewId pageList = 
    pageListGetCounter pageList << pageListIncCounter pageList

-- | Add page to page list.
pageListAddPage :: IORef PageList -> Page -> IO ()
pageListAddPage pageList page =
    modifyIORef pageList (\x -> setListAddNode x page)

-- | Remove page from page list.
pageListRemovePage :: IORef PageList -> Page -> IO ()
pageListRemovePage pageList page =
    modifyIORef pageList (\x -> setListRemoveNode x page)

-- | Get page.
pageListGetPage :: IORef PageList -> Int -> IO (Maybe Page)
pageListGetPage pageList id =
    (<<<=) (\x -> setListGetNode x (\y -> id == pageId y)) readIORef pageList

-- | Get page with name.
pageListGetPageWithName :: IORef PageList -> String -> IO (Maybe Page)
pageListGetPageWithName pageList name =
    (<=<) (\x -> setListGetNodeM x ((<<<=) (== name) pageGetName)) readIORef pageList

-- | Get page list.
pageListGetList :: IORef PageList -> IO [Page]
pageListGetList =
    (<<<=) setListGetList readIORef

-- | Clone page list.
pageListClone :: IORef PageList -> IO (IORef PageList)
pageListClone pageList = do
  newPageList <- pageListNew
  mapM_ (\x -> pageClone x >>= pageListAddPage newPageList) =<< pageListGetList pageList
  return newPageList

-- | Get clone page.
pageListGetClonePage :: IORefObject -> Page -> IO (Maybe Page)
pageListGetClonePage ioRefObject = 
    pageListGetPage (ioRefObjectPageList ioRefObject) . pageId
