module Manatee.UI.TabGroup where

import Data.IORef
import Data.Sequence (Seq)
import Data.List
import Distribution.Simple.Utils
import Control.Monad

import Graphics.UI.Gtk

import Manatee.UI.Page 
import Manatee.Utils.Seq
import Manatee.Utils.Generic hiding (swap)
import Manatee.Utils.Gtk.Gtk
import Manatee.IORefObject

import qualified Data.Sequence as S
import qualified Data.Foldable as F

data TabGroup = 
    TabGroup {tabGroupNotebook          :: Notebook
             ,tabGroupPageList          :: IORef PageList
             ,tabGroupPageIdSequence    :: IORef (Seq Int)} 

-- | Create tab group with page list.
tabGroupNew :: IORef PageList -> Maybe [Int] -> IO TabGroup
tabGroupNew pageList pageIdList = do
  -- Clone page list.
  plc <- pageListClone pageList

  -- Create tabGroup.
  nb  <- notebookNew
  seq <- newIORef S.empty
  let tabGroup = TabGroup nb plc seq

  -- Add page to notebook before setup current page of notebook.
  -- Otherwise, `notebookCurrentPage` can't work.
  tabGroupAppendPageList tabGroup plc pageIdList

  -- Setup notebook property.
  set nb [notebookTabPos        := PosTop
         ,notebookShowTabs      := True
         ,notebookScrollable    := True
         ,notebookEnablePopup   := True
         ,notebookCurrentPage   := 0]

  -- Return tabGroup.
  return tabGroup

-- | Clone tab group.
tabGroupClone :: TabGroup -> IO TabGroup
tabGroupClone tabGroup = do
  let pageList = tabGroupPageList tabGroup
  pageIdList <- (<<<=) (Just . F.toList) (readIORef . tabGroupPageIdSequence) tabGroup
  tabGroupNew pageList pageIdList

-- | Append page to notebook.
tabGroupAppendPage :: TabGroup -> PageInsertPosition -> Int -> Page -> IO ()
tabGroupAppendPage tabGroup position id page = do
  -- Get insert id.
  insertId <- case position of
               PStart -> tabGroupFirstIndex tabGroup
               PEnd   -> (<<<=) succ tabGroupLastIndex tabGroup
               PLeft  -> return id
               PRight -> return $ succ id

  -- Insert page.
  tabGroupInsertPage tabGroup page insertId

-- | Append page list to notebook.
tabGroupAppendPageList :: TabGroup -> IORef PageList -> Maybe [Int] -> IO ()
tabGroupAppendPageList tabGroup pageList = 
    applyMaybeTM_ (mapM_ (\i -> applyMaybeTM_ (tabGroupAppendPage tabGroup PEnd 0)
                                =<< pageListGetPage pageList i)) 

-- | Insert notebook with id.
tabGroupInsertPage :: TabGroup -> Page -> Int  -> IO ()
tabGroupInsertPage tabGroup page insertId = do
  let notebook  = tabGroupNotebook tabGroup
      box       = pageBox page
      sequencen = tabGroupPageIdSequence tabGroup
      pId       = pageId page
  name <- pageGetName page

  -- Show child widget before add,
  -- otherwise notebook can't setup current page.
  -- This is a problem of GTK C code.
  widgetShowAll box

  -- Insert page to nobeook.
  notebookInsertPage notebook box (fromUTF8 name) insertId
  notebookSetTabReorderable notebook box True -- make tab can reorderable

  -- Update page id sequence.
  modifyIORef sequencen (insertAt insertId pId)

-- Append page to notebook and add new page to page list.
tabGroupAddPage :: TabGroup -> PageInsertPosition -> Int -> Page -> IO ()
tabGroupAddPage tabGroup position id page = do
  tabGroupAppendPage tabGroup position id page
  pageListAddPage (tabGroupPageList tabGroup) page

-- | Remove all page.
tabGroupRemoveAllPage :: TabGroup -> IO ()
tabGroupRemoveAllPage = containerRemoveAll . tabGroupNotebook

-- | Remove page.
tabGroupRemovePage :: TabGroup -> Page -> IO ()
tabGroupRemovePage tabGroup page = 
    -- Remove page that have same page id from page list.
    applyMaybeTM_ (\p -> do
                    -- Remove page that have same page id from all views.
                    pageListRemovePage (tabGroupPageList tabGroup) p

                    -- Remove page from notebook.
                    applyMaybeTM_ (\i -> do
                                    -- Remove page from notebook.
                                    notebookRemovePage (tabGroupNotebook tabGroup) i

                                    -- Remove page from sequence.
                                    modifyIORef (tabGroupPageIdSequence tabGroup) (deleteAt i))
                                       =<< tabGroupGetPageSequnceId tabGroup page)
        =<< tabGroupGetClonePage tabGroup page

-- | Focus page.
tabGroupFocusPage :: TabGroup -> Page -> IO ()
tabGroupFocusPage tabGroup page = 
    applyMaybeTM_ (tabGroupSelectPage tabGroup)
        =<< tabGroupGetPageSequnceId tabGroup page

-- | Get current page.
tabGroupGetCurrentPage :: TabGroup -> IO (Maybe Page)
tabGroupGetCurrentPage tabGroup = 
    tabGroupGetPageSequncePage tabGroup
        =<< tabGroupCurrentIndex tabGroup

-- | Select next page.
tabGroupSelectNextPage :: TabGroup -> IORefObject -> IO ()
tabGroupSelectNextPage tabGroup ioRefObject = 
    ifM (tabGroupAtEnd tabGroup)
            (tabGroupSelectFirstPage tabGroup ioRefObject)
            (notebookNextPage $ tabGroupNotebook tabGroup)

-- | Select previous page.
tabGroupSelectPrevPage :: TabGroup -> IORefObject -> IO ()
tabGroupSelectPrevPage tabGroup ioRefObject = 
    ifM (tabGroupAtStart tabGroup)
            (tabGroupSelectLastPage tabGroup ioRefObject)
            (notebookPrevPage $ tabGroupNotebook tabGroup)

-- | Select first page.
tabGroupSelectFirstPage :: TabGroup -> IORefObject -> IO ()
tabGroupSelectFirstPage tabGroup _ = 
    tabGroupSelectPage tabGroup
        =<< tabGroupFirstIndex tabGroup

-- | Select last page.
tabGroupSelectLastPage :: TabGroup -> IORefObject -> IO ()
tabGroupSelectLastPage tabGroup _ = 
    tabGroupSelectPage tabGroup
        =<< tabGroupLastIndex tabGroup

-- | Select page with given index.
tabGroupSelectPage :: TabGroup -> Int -> IO ()
tabGroupSelectPage = notebookSetCurrentPage . tabGroupNotebook

-- | At start.
tabGroupAtStart :: TabGroup -> IO Bool
tabGroupAtStart tabGroup = 
    liftM2 (==) (tabGroupCurrentIndex tabGroup) (tabGroupFirstIndex tabGroup)

-- | At End.
tabGroupAtEnd :: TabGroup -> IO Bool
tabGroupAtEnd tabGroup = 
    liftM2 (==) (tabGroupCurrentIndex tabGroup) (tabGroupLastIndex tabGroup)
  
-- | Get tab first index.
tabGroupFirstIndex :: TabGroup -> IO Int
tabGroupFirstIndex _ = return 0

-- | Get tab last index.
tabGroupLastIndex :: TabGroup -> IO Int
tabGroupLastIndex = 
  (<<<=) pred tabGroupSize

-- | Get current index.
tabGroupCurrentIndex :: TabGroup -> IO Int
tabGroupCurrentIndex = notebookGetCurrentPage . tabGroupNotebook

-- | Get tab group size.
tabGroupSize :: TabGroup -> IO Int
tabGroupSize = notebookGetNPages . tabGroupNotebook

-- | Get sequencen id with given page.
tabGroupGetPageSequnceId :: TabGroup -> Page -> IO (Maybe Int)
tabGroupGetPageSequnceId tabGroup page = 
    (<<<=) (findIndex (== pageId page) . F.toList) (readIORef . tabGroupPageIdSequence) tabGroup

-- | Get page with given seuqnce id.
tabGroupGetPageSequncePage :: TabGroup -> Int -> IO (Maybe Page) 
tabGroupGetPageSequncePage tabGroup sequenceId = 
    -- Get page.
    applyMaybeTM (pageListGetPage (tabGroupPageList tabGroup)) 
        -- Get page id.
        =<< (<<<=) (\p -> F.toList p `maybeIndex` sequenceId) (readIORef . tabGroupPageIdSequence) tabGroup 

-- | Get clone page.
tabGroupGetClonePage :: TabGroup -> Page -> IO (Maybe Page)
tabGroupGetClonePage tabGroup = 
    pageListGetPage (tabGroupPageList tabGroup) . pageId   

-- | Set page name.
tabGroupSetPageName :: TabGroup -> Page -> String -> IO ()
tabGroupSetPageName tabGroup page name = 
    applyMaybeTM_ (\clonePage -> do
                     -- Set page name.
                     pageSetName clonePage name

                     -- Set notebook label name.
                     notebookSetTabLabelText (tabGroupNotebook tabGroup) (pageBox clonePage) name
                  ) =<< tabGroupGetClonePage tabGroup page

-- | Apply focus tab group.
tabGroupApplyFocus :: (TabGroup -> IORefObject -> IO ()) -> IORefObject -> IO ()
tabGroupApplyFocus f ioRefObject =
    applyMaybeTM_ (\t -> f t ioRefObject)
        =<< ioRefObjectGetFocusTabGroup ioRefObject

-- | Move current page to left.
-- If current page is first page, keep stay.
tabGroupMovePageToLeft :: TabGroup -> IORefObject -> IO ()
tabGroupMovePageToLeft tabGroup ioRefObject = 
    unlessM (tabGroupAtStart tabGroup) $ 
            applyMaybeTM_ (\p -> do       
                             id <- tabGroupCurrentIndex tabGroup
                             ioRefObjectMovePage p (id - 1) ioRefObject
                          ) =<< tabGroupGetCurrentPage tabGroup

-- | Move current page to right.
-- If current page is first page, keep stay.
tabGroupMovePageToRight :: TabGroup -> IORefObject -> IO ()
tabGroupMovePageToRight tabGroup ioRefObject = 
    unlessM (tabGroupAtEnd tabGroup) $ 
            applyMaybeTM_ (\p -> do
                             id <- tabGroupCurrentIndex tabGroup
                             ioRefObjectMovePage p (id + 1) ioRefObject
                             ) =<< tabGroupGetCurrentPage tabGroup

-- | Move current page to begin.
-- If current page is begin page, keep stay.
tabGroupMovePageToBegin :: TabGroup -> IORefObject -> IO ()
tabGroupMovePageToBegin tabGroup ioRefObject = 
    unlessM (tabGroupAtStart tabGroup) $ 
            applyMaybeTM_ (\p -> do
                             id <- tabGroupFirstIndex tabGroup
                             ioRefObjectMovePage p id ioRefObject
                          ) =<< tabGroupGetCurrentPage tabGroup

-- | Move current page to end.
-- If current page is end page, keep stay.
tabGroupMovePageToEnd :: TabGroup -> IORefObject -> IO ()
tabGroupMovePageToEnd tabGroup ioRefObject = 
    unlessM (tabGroupAtEnd tabGroup) $ 
            applyMaybeTM_ (\p -> do
                             id <- tabGroupLastIndex tabGroup
                             ioRefObjectMovePage p id ioRefObject
                          ) =<< tabGroupGetCurrentPage tabGroup
