module Manatee.UI.View where

import Data.Maybe
import Data.List
import Data.Function
import Text.Printf
import Data.IORef
import Control.Monad
import Data.Ord

import Graphics.UI.Gtk hiding (on, get)
import Graphics.UI.Gtk.Abstract.Container
import Graphics.UI.Gtk.Abstract.Paned

import Manatee.UI.ViewNode
import Manatee.UI.Page
import Manatee.UI.TabGroup
import Manatee.Data.ListZipper (ListZipper)
import Manatee.IORefObject
import Manatee.Utils.Generic
import Manatee.PageView.PageView

import qualified Manatee.Data.ListZipper as LZ

data View =
    View {viewNode     :: ViewNode
         ,viewTabGroup :: TabGroup}

instance Eq View where
    (==) = (==) `on` viewNode

instance Ord View where
    compare = comparing viewNode

instance Show View where
    show = printf "<#View %s>" . show . viewNodeId . viewNode

-- | Init view.
viewInit :: IORefObject -> IO ()
viewInit ioRefObject = do
  -- Get view listzipper.
  vlz <- (<<<=) viewListZipper (readIORef . ioRefObjectViewList) ioRefObject

  -- Just create first view when view list is empty.
  when (LZ.isEmpty vlz) $ 
       viewRootNew ioRefObject >> return ()

-- | New root view.
viewRootNew :: IORefObject -> IO View
viewRootNew ioRefObject = do
  -- Create new view.
  let viewList = ioRefObjectViewList ioRefObject
  view <- viewNew TNodeRoot DTop Nothing Nothing ioRefObject

  -- Focus view.
  viewListFocus viewList view

  return view

-- | Create new view.
viewNew :: ViewNodeType -> ViewNodeDirection -> Maybe ViewNode -> Maybe View -> IORefObject -> IO View
viewNew vnType direction parentNode viewParent ioRefObject = do
  -- New view node.
  vn <- viewNodeNew parentNode vnType direction ioRefObject

  viewNewInternal vn viewParent ioRefObject

-- | Create new view internal.
viewNewInternal :: ViewNode -> Maybe View -> IORefObject -> IO View
viewNewInternal vn viewParent ioRefObject = do
  let viewList = ioRefObjectViewList ioRefObject
      pageList = ioRefObjectPageList ioRefObject

  -- New tabgroup.
  tabGroup <- case applyMaybe viewTabGroup viewParent of
               -- Clone tabgroup for split view.
               Just tg -> tabGroupClone tg
               -- New tabgroup for root view.
               Nothing -> tabGroupNew pageList Nothing

  -- Focus page.
  applyMaybeTM_ (tabGroupFocusPage tabGroup) 
      =<< applyMaybeTM viewGetCurrentPage viewParent

  -- New view.
  let view = View vn tabGroup

  -- Add tabGroup to view.
  viewGetTopContainer view `containerAdd` tabGroupNotebook tabGroup

  -- Add view to view list.
  vnType <- readIORef $ viewNodeType vn
  viewListAddView viewList view vnType

  -- Add action handle callback function.
  -- This step is NECESSERY for focus widget in view.
  viewHandleAction view viewList

  return view

-- | Handle user action.
viewHandleAction :: View -> IORef ViewList -> IO ()
viewHandleAction view viewList = do
  -- Setup focus view when swtich page in notebook.
  let notebook = tabGroupNotebook $ viewTabGroup view
  notebook `onSwitchPage` (\_ -> viewListFocus viewList view)

  return ()

-- | Get top view container.
viewGetTopContainer :: View -> Paned
viewGetTopContainer = viewNodePaned . viewNode

-- | Split view with indicate direction.
viewSplit :: View -> ViewNodeDirection -> IORefObject -> IO (View, View) 
viewSplit viewParent direction ioRefObject = do
  let viewList = ioRefObjectViewList ioRefObject
  
  -- Get split node.
  parentNode <- viewNodeGetSplitContainer (viewNode viewParent) direction ioRefObject
  let childContainner = viewNodePaned parentNode

  -- Create view children.
  viewChild1 <- viewNew TNodeLeft  direction (Just parentNode) (Just viewParent) ioRefObject
  viewChild2 <- viewNew TNodeRight direction (Just parentNode) (Just viewParent) ioRefObject

  -- Remove old view from view list.
  viewListRemoveView viewList

  -- Show all widgets, this step is NECESSERY!
  widgetShowAll childContainner

  -- Focus view.
  viewListFocus viewList viewChild1

  -- Return child view along with split direction.
  if viewNodeIsFirstDirection direction
     then return (viewChild1, viewChild2)
     else return (viewChild2, viewChild1)

-- | Split view vertically.
viewSplitVertically :: View -> IORefObject -> IO ()  
viewSplitVertically view ioRefObject = 
    viewSplit view DTop ioRefObject >> return () 

-- | Split view hortizontally.
viewSplitHortizontally :: View -> IORefObject -> IO ()
viewSplitHortizontally view ioRefObject = 
    viewSplit view DLeft ioRefObject >> return ()

-- | Remove view.
viewRemove :: View -> IORefObject -> IO ()
viewRemove view ioRefObject = do
  let viewList = ioRefObjectViewList ioRefObject
  -- Get view list size.
  size <- viewListGetSize viewList

  if size == 1
     -- Don't delete view if just only one.
     then putStrLn "Just only one view, you can't delete it."
     -- Otherwise remove view.
     else do 
       -- Remove view node.
       viewNodeRemove (viewNode view) ioRefObject True

       -- Remove from viewList.
       viewListRemoveView viewList 

-- | Remove others views except current view.
viewRemoveOthers :: View -> IORefObject -> IO ()
viewRemoveOthers view ioRefObject =  do
  let viewList = ioRefObjectViewList ioRefObject
  listSize <- viewListSize viewList

  case listSize of
    -- Don't delete view when haven't or just only one view exist.
    0 -> putStrLn "Haven't any view exist."
    1 -> putStrLn "Haven't others view exist."
    -- Otherwise delete others view except current one.
    _ -> do
      -- Remove others view nodes.
      viewNodeRemoveOthers (viewNode view) ioRefObject

      -- Remove othersw view from view list.
      viewListRemoveOthersView viewList 

      -- Focus view.
      viewListFocus viewList view

      return () 

-- | Get current page sequencen id.
viewGetCurrentPageSequenceId :: View -> IO (Maybe Int)
viewGetCurrentPageSequenceId view = 
    applyMaybeTM (tabGroupGetPageSequnceId (viewTabGroup view)) 
        =<< viewGetCurrentPage view

-- | Get view page list.
viewGetPageList :: View -> IO [Page]
viewGetPageList = pageListGetList . tabGroupPageList . viewTabGroup

-- | Get view current page.
viewGetCurrentPage :: View -> IO (Maybe Page)
viewGetCurrentPage = tabGroupGetCurrentPage . viewTabGroup

-- | Focus view buffer.
viewFocusBuffer :: View -> IO ()
viewFocusBuffer view = 
    applyMaybeTM_ (pageApplyBuffer pageViewFocus)
        =<< viewGetCurrentPage view

-- | Apply focus view.
viewApplyFocus :: (View -> IORefObject -> IO ()) -> IORefObject -> IO ()
viewApplyFocus f ioRefObject =
    applyMaybeTM_ (\v -> f v ioRefObject)
        =<< ioRefObjectGetFocusView ioRefObject

-- | Apply focus view.
viewApplyFocus' :: (View -> IO (Maybe a)) -> IORefObject -> IO (Maybe a)
viewApplyFocus' f ioRefObject =
    applyMaybeTM f
        =<< ioRefObjectGetFocusView ioRefObject

-- | Apply view with speical id.
viewApplyWithId :: (View -> IO ()) -> Int -> IORefObject -> IO ()
viewApplyWithId f id ioRefObject =
    applyMaybeTM_ f =<< viewListGetView id ioRefObject

data ViewList =
    ViewList {viewListZipper    :: ListZipper View}

data ViewListSelectDirection = VLeft 
                             | VRight 
                               deriving (Eq, Show, Read)

-- | Create view list.
viewListNew :: IO (IORef ViewList)
viewListNew = newIORef $ ViewList LZ.empty

-- | Add view to view list.
viewListAddView :: IORef ViewList -> View -> ViewNodeType -> IO ()
viewListAddView viewList view vnType = 
    viewListApply2 viewList view $ 
                   case vnType of
                     TNodeLeft -> LZ.insertLeft
                     _         -> LZ.insertRight

-- | Remove view from view list.
viewListRemoveView :: IORef ViewList -> IO ()
viewListRemoveView viewList = do
  nextFocus <- (<<<=) (LZ.getRightCircular . viewListZipper) readIORef viewList

  -- Remove from view list.
  viewList `viewListApplyM1` LZ.delete

  -- Move to next view if have next focus view. 
  applyMaybeTM_ (viewListFocus viewList) nextFocus

-- | Remove others view except current view from view list.
viewListRemoveOthersView :: IORef ViewList -> IO ()
viewListRemoveOthersView = (`viewListApplyM1` LZ.deleteOthers)

-- | Focus view.
viewListFocus :: IORef ViewList -> View -> IO ()
viewListFocus viewList view = do
  -- Focus current view in list.
  viewListApplyM2 viewList view LZ.focusNode

  -- Focus view buffer.
  viewFocusBuffer view

-- | Get current focus view.
viewListGetFocusView :: IORef ViewList -> IO (Maybe View)
viewListGetFocusView = 
    (<<<=) (LZ.getCurrent . viewListZipper) readIORef 

-- | Ge tview with speical id.
viewListGetView :: Int -> IORefObject -> IO (Maybe View)
viewListGetView id ioRefObject = do
  vList <- viewListGetList ioRefObject
  return $ find (\v -> viewNodeId (viewNode v) == id) vList

-- | Select view with given direction.
viewListSelect :: ViewListSelectDirection -> Bool -> IORef ViewList -> IO ()
viewListSelect direction circular viewList = 
  -- Move next focus if have it.
  applyMaybeTM_ (viewListFocus viewList) =<< (<<<=) (selectF . viewListZipper) readIORef viewList 

  -- Get next focus function.
  where selectF = 
            case direction of
              VLeft  -> if circular
                          then LZ.getLeftCircular
                          else LZ.getLeft
              VRight -> if circular
                          then LZ.getRightCircular
                          else LZ.getRight 

-- | Next view in view list.
viewListNext :: IORefObject -> IO ()
viewListNext = viewListSelect VRight False . ioRefObjectViewList

-- | Previous view in view list.
viewListPrev :: IORefObject -> IO ()
viewListPrev = viewListSelect VLeft False . ioRefObjectViewList

-- | Next view circular in view list.
viewListNextCircular :: IORefObject -> IO ()
viewListNextCircular = viewListSelect VRight True . ioRefObjectViewList

-- | Previous view circular in view list.
viewListPrevCircular :: IORefObject -> IO ()
viewListPrevCircular = viewListSelect VLeft True . ioRefObjectViewList

-- | Return view list size.
viewListSize :: IORef ViewList -> IO Int
viewListSize  = (<<<=) (LZ.length . viewListZipper) readIORef

-- | Get list.
viewListGetList :: IORefObject -> IO [View]
viewListGetList = (<<<=) (LZ.toList . viewListZipper) readIORef . ioRefObjectViewList

-- | Show view list.
viewListShow :: IORefObject -> IO ()
viewListShow ioRefObject = print . show =<< viewListGetList ioRefObject

-- | Apply view list.
viewListApplyM1 :: IORef ViewList -> (ListZipper View -> Maybe (ListZipper View)) -> IO ()
viewListApplyM1 v f = 
    modifyIORef v (\x -> case f $ viewListZipper x of
                          Just vlz -> x {viewListZipper = vlz}
                          Nothing  -> x)

-- | Apply view list.
viewListApplyM2 :: IORef ViewList -> View -> (View -> ListZipper View -> Maybe (ListZipper View)) -> IO ()
viewListApplyM2 vl v f = 
    modifyIORef vl (\x -> case f v $ viewListZipper x of
                           Just vlz -> x {viewListZipper = vlz}
                           Nothing  -> x)

-- | Apply view list.
viewListApply2 :: IORef ViewList -> View -> (View -> ListZipper View -> ListZipper View) -> IO ()
viewListApply2 vl v f = 
    modifyIORef vl (\x -> x {viewListZipper = f v $ viewListZipper x})

-- | Get view list size.
viewListGetSize :: IORef ViewList -> IO Int
viewListGetSize = (<<<=) (LZ.length . viewListZipper) readIORef

-- | Apply view.
viewListApplyView :: (View -> IO ()) -> IORefObject -> IO ()
viewListApplyView f ioRefObject = mapM_ f =<< viewListGetList ioRefObject
