{-# LANGUAGE ExistentialQuantification #-}
module Manatee.UI.ViewNode where

import Control.Applicative
import Data.IORef
import Data.Maybe 
import Text.Printf
import Control.Monad
import Data.Function
import Data.Ord

import Graphics.UI.Gtk hiding (on, get)
import Graphics.UI.Gtk.Abstract.Paned
import Graphics.UI.Gtk.Layout.HPaned
import Graphics.UI.Gtk.Layout.VPaned

import Manatee.IORefObject
import Manatee.Utils.Generic
import Manatee.Utils.Gtk.Gtk
import Manatee.Data.SetList

-- If you create N View, you need (2N - 1) ViewNode for suppot it.
data ViewNode =
    ViewNode {viewNodeId                :: Int         
             ,viewNodePaned             :: Paned
             ,viewNodeParentId          :: IORef (Maybe Int)
             ,viewNodeChildLeftId       :: IORef (Maybe Int)
             ,viewNodeChildRightId      :: IORef (Maybe Int)
             ,viewNodeType              :: IORef ViewNodeType
             ,viewNodeDirection         :: ViewNodeDirection}

data ViewNodeType = TNodeLeft
                  | TNodeRight
                  | TNodeRoot
                    deriving (Eq, Show, Read)

data ViewNodeDirection = DTop
                       | DBottom
                       | DLeft
                       | DRight
                         deriving (Eq, Show, Read)

instance Eq ViewNode where
    (==) = (==) `on` viewNodeId

instance Ord ViewNode where
    compare = comparing viewNodeId

-- | Create new view node.
viewNodeNew :: Maybe ViewNode -> ViewNodeType -> ViewNodeDirection -> IORefObject -> IO ViewNode
viewNodeNew parentNode = 
    viewNodeNewInternal Nothing (applyMaybe viewNodeId parentNode) Nothing Nothing 

-- | Internal fucntion for create view node.  
viewNodeNewInternal :: Maybe Int -> Maybe Int -> Maybe Int -> Maybe Int -> ViewNodeType -> ViewNodeDirection -> IORefObject -> IO ViewNode
viewNodeNewInternal vId pId clId crId typ direction ioRefObject = do
  let viewNodeList = ioRefObjectViewNodeList ioRefObject

  -- Init value.
  id <- case vId of
         -- Just return speical id.
         Just i  -> return i
         -- Otherwise create new id.
         Nothing -> viewNodeListGetNewId viewNodeList
  paned    <- viewNodePanedNew direction
  parentId <- newIORef pId
  leftId   <- newIORef clId
  rightId  <- newIORef crId
  vType    <- newIORef typ
    
  -- Create new view node.
  let viewNode = ViewNode id paned parentId leftId rightId vType direction

  -- Connect current node parent.
  viewNodeConnectToParent viewNode ioRefObject

  -- Add to view node list.
  viewNodeListAddNode viewNodeList viewNode
  
  return viewNode

-- | Remvoe view node. 
viewNodeRemove :: ViewNode -> IORefObject -> Bool -> IO ()
viewNodeRemove viewNode ioRefObject removeOtioseNode = do
  let viewNodeList = ioRefObjectViewNodeList ioRefObject
  vType <- readIORef $ viewNodeType viewNode

  -- Remove view node from view node list first.
  viewNodeListRemoveNode viewNodeList viewNode

  -- Remove view node paned.
  viewNodePanedRemove viewNode ioRefObject

  -- Clean otiose nodes.
  case vType of
    TNodeRoot -> return ()
    _         -> 
        applyMaybeTM_ (\n -> do
                         -- Remove current node from parent node.
                         viewNodeRemoveFromParentNode viewNode n

                         -- Remove otiose node.
                         when removeOtioseNode $ do                             
                           -- Is have child node?
                           haveChildNode <- viewNodeIsHaveChildNode n

                           -- Clean otiose node when haven't children node under current node.
                           unless haveChildNode $ viewNodeRemove n ioRefObject True)
            =<< viewNodeGetParentNode viewNode ioRefObject

-- | Remove others view ndoes except current view node.
viewNodeRemoveOthers :: ViewNode -> IORefObject -> IO ()
viewNodeRemoveOthers viewNode ioRefObject = do
  let viewNodeList = ioRefObjectViewNodeList ioRefObject
      viewNodeRootContainer = ioRefObjectRootWindow ioRefObject

  -- Change view node to root view node.
  viewNodeChangeToRoot viewNode ioRefObject
  containerRemoveAll viewNodeRootContainer                    -- remove all child from root window
  viewNodeRootContainer `containerAdd` viewNodePaned viewNode -- add view node to root window

  -- Remove others view node from view node list.
  viewNodeListRemoveOthersNode viewNodeList viewNode 

  return ()

-- | Get parent node of current node.
viewNodeGetParentNode :: ViewNode -> IORefObject -> IO (Maybe ViewNode)
viewNodeGetParentNode viewNode ioRefObject = 
  applyMaybeTM (viewNodeListGetNode (ioRefObjectViewNodeList ioRefObject))
      =<< readIORef (viewNodeParentId viewNode)

-- | Get split container for child view node.
viewNodeGetSplitContainer :: ViewNode -> ViewNodeDirection -> IORefObject -> IO ViewNode
viewNodeGetSplitContainer viewNode direction ioRefObject = 
    ifM (viewNodeIsSameSplitDirection (viewNodeDirection viewNode) direction)
             -- Return current view node if child node is same split direction.
             (do
               containerRemoveAll $ viewNodePaned viewNode
               return viewNode)
             -- Otherwise, create new container node from parent node of current node.
             (do
               -- Remove current node first.
               viewNodeRemove viewNode ioRefObject False

               -- Create new view node from parent node.
               parentNode <- viewNodeGetParentNode viewNode ioRefObject
               vType      <- readIORef $ viewNodeType viewNode
               viewNodeNew parentNode vType direction ioRefObject)

-- | Change view node to root node.
viewNodeChangeToRoot :: ViewNode -> IORefObject -> IO ()
viewNodeChangeToRoot viewNode ioRefObject = do
  -- Remove view paned from parent container.
  viewNodePanedRemove viewNode ioRefObject   

  -- Change view node attribute.
  writeIORef (viewNodeParentId viewNode) Nothing
  writeIORef (viewNodeChildLeftId viewNode) Nothing
  writeIORef (viewNodeChildRightId viewNode) Nothing
  writeIORef (viewNodeType viewNode) TNodeRoot 

-- | Add view node paned to parent container.
viewNodeConnectToParent :: ViewNode -> IORefObject -> IO ()
viewNodeConnectToParent viewNode ioRefObject = do
  let vPaned = viewNodePaned viewNode
  vType <- readIORef $ viewNodeType viewNode
  case vType of
    -- Add view node to root window if it is root node.
    TNodeRoot -> ioRefObjectRootWindow ioRefObject `containerAdd` vPaned
    -- Otherwise, add to other view node.
    _         -> 
        applyMaybeTM_ (\n -> do
                        -- Add current node to other view node.
                        let vParentPaned = viewNodePaned n
                            id = viewNodeId viewNode
                        case vType of
                          TNodeLeft  -> do
                                     panedPack1 vParentPaned vPaned True True
                                     writeIORef (viewNodeChildLeftId n) (Just id)
                          TNodeRight -> do
                                     panedPack2 vParentPaned vPaned True True
                                     writeIORef (viewNodeChildRightId n) (Just id)
                      ) =<< viewNodeGetParentNode viewNode ioRefObject

-- | Get view node paned with indicate split direction.
viewNodePanedNew :: ViewNodeDirection -> IO Paned
viewNodePanedNew direction
             | viewNodeIsVerticallyDirection direction 
                 = castToPaned <$> vPanedNew
             | otherwise                               
                 = castToPaned <$> hPanedNew

-- | Remove node paned from parent container.
viewNodePanedRemove :: ViewNode -> IORefObject -> IO ()
viewNodePanedRemove viewNode ioRefObject = do
  let vPaned = viewNodePaned viewNode
  vType <- readIORef $ viewNodeType viewNode

  case vType of
    -- Just remove from root window if current node is root node.
    TNodeRoot -> ioRefObjectRootWindow ioRefObject `containerRemove` vPaned
    -- Otherwise, remove from parent view node.
    _         -> 
        applyMaybeTM_ (\n -> 
                       -- Remove paned from parent node paned.
                       viewNodePaned n `containerRemove` vPaned)
            =<< viewNodeGetParentNode viewNode ioRefObject

-- | Is haven't any child node.
viewNodeIsHaveChildNode :: ViewNode -> IO Bool
viewNodeIsHaveChildNode viewNode = do
  childLeftId  <- readIORef $ viewNodeChildLeftId viewNode
  childRightId <- readIORef $ viewNodeChildRightId viewNode
  
  return $ childLeftId /= Nothing || childRightId /= Nothing

-- | Remove from parent node.
viewNodeRemoveFromParentNode :: ViewNode -> ViewNode -> IO ()  
viewNodeRemoveFromParentNode viewNode n = do
  vnType <- readIORef $ viewNodeType viewNode
  case vnType of
    TNodeLeft  -> writeIORef (viewNodeChildLeftId n) Nothing
    TNodeRight -> writeIORef (viewNodeChildRightId n) Nothing
    _          -> return ()

-- | Return true if two view nodes is same split directon. 
viewNodeIsSameSplitDirection :: ViewNodeDirection -> ViewNodeDirection -> IO Bool
viewNodeIsSameSplitDirection dir1 dir2 = return $ f dir1 == f dir2
    where
      f = viewNodeIsVerticallyDirection

-- | Whether split direction is vertically.
viewNodeIsVerticallyDirection :: ViewNodeDirection -> Bool
viewNodeIsVerticallyDirection = (`elem` [DTop, DBottom])

-- | Whether direction is `DTop` or `DLeft`.
viewNodeIsFirstDirection :: ViewNodeDirection -> Bool
viewNodeIsFirstDirection = (`elem` [DTop, DLeft])

-- | Show view node.
viewNodeShow :: ViewNode -> IO ()
viewNodeShow viewNode = do
  let id = viewNodeId viewNode
  parentId <- readIORef $ viewNodeParentId viewNode
  leftId   <- readIORef $ viewNodeChildLeftId viewNode
  rightId  <- readIORef $ viewNodeChildRightId viewNode
  printf "Node (%s) <ParentId: %s, LeftCId: %s, RightCId: %s>\n" (show id) (show parentId) (show leftId) (show rightId)

type ViewNodeList = SetList ViewNode

-- | New view ndoe.
viewNodeListNew :: IO (IORef ViewNodeList)
viewNodeListNew = newIORef setListNew

-- | Get counter.
viewNodeListGetCounter :: IORef ViewNodeList -> IO Int
viewNodeListGetCounter = 
    (<<<=) setListGetCounter readIORef

-- | Increase counter.
viewNodeListIncCounter :: IORef ViewNodeList -> IO ()
viewNodeListIncCounter viewNodeList = 
    modifyIORef viewNodeList setListIncCounter

-- | Get new id for create view ndoe.
viewNodeListGetNewId :: IORef ViewNodeList -> IO Int
viewNodeListGetNewId viewNodeList = 
    viewNodeListGetCounter viewNodeList << viewNodeListIncCounter viewNodeList

-- | Add view node to list.
viewNodeListAddNode :: IORef ViewNodeList -> ViewNode -> IO ()
viewNodeListAddNode viewNodeList viewNode = 
    modifyIORef viewNodeList (\x -> setListAddNode x viewNode)

-- | Remove view node from list.
viewNodeListRemoveNode :: IORef ViewNodeList -> ViewNode -> IO ()
viewNodeListRemoveNode viewNodeList viewNode = 
    modifyIORef viewNodeList (\x -> setListRemoveNode x viewNode)

-- | Remove others view ndoe except current view node from list.
viewNodeListRemoveOthersNode :: IORef ViewNodeList -> ViewNode -> IO ()
viewNodeListRemoveOthersNode viewNodeList viewNode = do
  -- Empty view node list set.
  viewNodeListEmptySet viewNodeList

  -- Add current view node.
  viewNodeListAddNode viewNodeList viewNode

-- | Get view ndoe with given id.
viewNodeListGetNode :: IORef ViewNodeList -> Int -> IO (Maybe ViewNode)
viewNodeListGetNode viewNodeList id = 
    (<<<=) (\x -> setListGetNode x (\y -> viewNodeId y == id)) readIORef viewNodeList

-- | Empty view node list set.
viewNodeListEmptySet :: IORef ViewNodeList -> IO ()
viewNodeListEmptySet viewNodeList =
    modifyIORef viewNodeList setListEmpty

-- | Show view node list.
viewNodeListShow :: IORef ViewNodeList -> IO ()
viewNodeListShow viewNodeList = 
    mapM_ viewNodeShow =<< (<<<=) setListGetList readIORef viewNodeList
