{-# LANGUAGE TransformListComp, FlexibleContexts #-}
module Manatee.Utils.Generic where

import Data.List
import Data.Array
import Data.Monoid
import Text.Regex.TDFA
import System.FilePath
import System.Directory
import System.IO
import System.Process
import Data.ByteString.Lazy.UTF8 (ByteString)
import Distribution.Simple.Utils

import Data.Maybe
import Data.Dynamic
import Data.Typeable
import Control.Arrow
import Control.Monad hiding (filterM, mapM)
import Data.Either

import qualified Control.Monad as CM
import qualified Data.ByteString.Lazy.UTF8 as U

-- | Indicate error or return a.
-- This function for replace `fromJust`, expression `fromJust x` is bad
-- when `x` is `Nothing`, so `maybeError` allowed you customize error
-- informatiron.
maybeError :: String -> Maybe a -> a
maybeError = fromMaybe . error

-- | Like find, but works with monadic computation instead of pure function.
-- In expression `find FUNCTION list`, if FUNCTION is "IO Bool", you can use
-- `findM FUNCTION list` to instead.
findM :: Monad m => (a -> m Bool) -> [a] -> m (Maybe a)
findM _ [] = return Nothing
findM f (x:xs) = ifM (f x) (return $ Just x) (findM f xs)

-- | Get type string of dynamic type.
dynTypeString :: Dynamic -> String
dynTypeString = tyConString . typeRepTyCon . dynTypeRep

-- | Type String.
typeString :: Typeable a => a -> String
typeString = tyConString . typeRepTyCon . typeOf

-- | Type of byte string.
typeByteString :: Typeable a => a -> ByteString
typeByteString = U.fromString . tyConString . typeRepTyCon . typeOf

-- | Apply two moand function with list. 
-- And return new monad tuples list.
applyM2 :: Monad m => [a] -> (a -> m b) -> (a -> m c) -> m [(b, c)]
applyM2 xs f g = forM xs (\x -> zipM' (f x) (g x))

-- | Apply two function with list.
-- And return new tuples list.
apply2 :: [a] -> (a -> b) -> (a -> c) -> [(b, c)]
apply2 xs f g = map (f &&& g) xs

-- | Return element of list with given index.
maybeIndex :: [a] -> Int -> Maybe a
maybeIndex [] _ = Nothing              
maybeIndex xs n
    | n < 0 || n >= length xs 
        = Nothing
    | otherwise               
        = listToMaybe . drop n $ xs

-- | Map left.
mapLeft :: (t1 -> a) -> Either t1 t -> Either a t
mapLeft _ (Right a) = Right a
mapLeft f (Left a)  = Left (f a)

-- | Map right.
mapRight :: (t1 -> b) -> Either t t1 -> Either t b
mapRight _ (Left a)  = Left a
mapRight f (Right a) = Right (f a)

-- | Apply either left.
applyEitherLeft :: (a -> c) -> (Either a b) -> Maybe c
applyEitherLeft f (Left l)  = Just $ f l
applyEitherLeft _ (Right _) = Nothing

-- | Apply either right.
applyEitherRight :: (b -> c) -> (Either a b) -> Maybe c
applyEitherRight _ (Left _)  = Nothing
applyEitherRight f (Right r) = Just $ f r

-- | Intersect element with list from end, don't including front or middle element.
-- Return null if haven't intersection element.
intersectEnd :: Ord a => [a] -> [a] -> [a]
intersectEnd xs ys = reverse $ intersectFront (reverse xs) (reverse ys)

-- | Intersect element with list from front, don't including end or middle element.
-- Return null if haven't intersection element.
intersectFront :: Ord a => [a] -> [a] -> [a]
intersectFront xs ys = [x | (x,y) <- zip xs ys, then takeWhile by x == y]

-- | Delay list with given start index.
delay :: Int -> [a] -> [Int]
delay n (_:xs) = n : delay (n + 1) xs
delay _ _      = []

-- | Index of list
listIndex :: [a] -> [Int]
listIndex = delay 0

-- | Pair with list index.
pairPred :: [a] -> [(a, Int)]
pairPred xs = zip xs $ listIndex xs

-- | Swap first and second in tuple.
swap :: (a,b) -> (b,a)
swap (a,b) = (b,a)

-- | If swap.
-- If true, swap tuple, otherwise keeping it.
ifSwap :: Bool -> (a, a) -> (a, a)
ifSwap b x 
    | b = swap x
    | otherwise = x

-- | Duplicat current value with tuple.
dup :: a -> (a, a)
dup a = (a,a)

-- | Different two list, and two lists must have same length.
-- otherwise throw a error.
different :: Ord a => [a] -> [a] -> [a]
different (x:xs) (y:ys) 
    | x == y    = different xs ys
    | otherwise = x : different xs ys
different [] [] = []
different _ _   = error "different: lists not the same length!"

-- | Apply maybe and return Bool.
applyMaybeB :: Maybe a -> Bool -> Bool
applyMaybeB (Just _) b = b
applyMaybeB Nothing b = not b

-- | Apply maybe.
applyMaybe :: (a -> b) -> Maybe a -> Maybe b
applyMaybe = fmap

-- | Apply maybe.
applyMaybe_ :: Monad m => (a -> b) -> Maybe a -> m (Maybe b)
applyMaybe_  f = return . fmap f

-- | Apply maybe with monad.
applyMaybeM :: Monad m => (a -> m b) -> Maybe a -> m (Maybe b)
applyMaybeM = maybe (return Nothing) . (<=<) (return . Just)

-- | Apply maybe.
applyMaybeT :: (a -> Maybe b) -> Maybe a -> Maybe b
applyMaybeT = maybe Nothing

-- | Apply monad maybe.
applyMaybeTM :: Monad m => (a -> m (Maybe b)) -> Maybe a -> m (Maybe b)
applyMaybeTM = maybe (return Nothing)

-- | Apply maybe with io moand.
applyMaybeTM_ :: Monad m => (a -> m ()) -> Maybe a -> m ()
applyMaybeTM_ = maybe (return ()) 

-- | Partion list.
partitionM :: Monad m => (a -> m Bool) -> [a] -> m ([a], [a])
partitionM _ [] = return ([], [])
partitionM f (x:xs) = 
    liftM2 mappend (ifM (f x) (return ([x], [])) (return ([], [x]))) (partitionM f xs)

-- | Do action when list not empty.
unlessNull :: [a] -> IO () -> IO ()
unlessNull = unless . null

-- | Transform Int to Doube
integralToDouble :: Integral a => a -> Double
integralToDouble v = fromIntegral v :: Double

-- | Integer to Int.
integerToInt :: Integer -> Int
integerToInt t = fromInteger t :: Int

-- | Functor <<<=.
(<<<=) :: Functor f => (b -> c) -> (a -> f b) -> a -> f c
(<<<=) = (.) . fmap

-- | Functor =>>
(=>>>) :: Functor f => (a -> f b) -> (b -> c) -> a -> f c
(=>>>) = flip (fmap . fmap)

-- | Flip >>
(<<) :: Monad m => m b -> m a -> m b
(<<) = flip (>>)

(<<=<) :: Functor f => (b -> f c) -> (a -> b) -> a -> f c
(<<=<) = (.)

(>=>>) :: Functor f => (a -> b) -> (b -> f c) -> a -> f c
(>=>>) = flip (<<=<)

-- | Unlines except last one.
unlinesExceptLast :: [String] -> String
unlinesExceptLast [] = ""
unlinesExceptLast list = init $ unlines list

-- | Like LiftM2, but make function at midddle of two arguments.
liftM2' :: Monad m => m a -> (a -> b -> c) -> m b -> m c
liftM2' xs f = liftM2 f xs

liftM2_ :: Monad m => (a -> b -> c) -> m a -> m b -> m ()
liftM2_ f x y = liftM2 f x y >> return ()

-- | If monad.
ifM :: Monad m => m Bool -> m a -> m a -> m a
ifM f g h = f >>= \b -> if b then g else h

-- | When monad.
whenM :: Monad m => m Bool -> m () -> m ()
whenM f g = f >>= \b -> when b g

-- | Unless monad.
unlessM :: Monad m => m Bool -> m () -> m ()
unlessM f g = f >>= \b -> unless b g

-- | Zip'
zip' :: a -> b -> (a, b)
zip' a b = (a, b)

-- | ZipM'
zipM' :: Monad m => m a -> m b -> m (a, b)
zipM' = liftM2 zip' 

-- | Head monad list.
headM :: Monad m => m [a] -> m a
headM = liftM head

-- | Last monad list.
lastM :: Monad m => m [a] -> m a
lastM = liftM last

-- | FirstMap.
firstMap :: (b -> c) -> [(b, d)] -> [(c, d)]
firstMap = map . first

-- | ConcatM.
concatM :: Monad m => m [a] -> m [a] -> m [a]
concatM = liftM2 (++) 

-- | Arrow with duplicat.
dupA :: Arrow a => a b (b, b)
dupA = arr dup

-- | Arrow with curry.
curryA :: Arrow a => a ((b, c) -> d) (b -> c -> d)
curryA = arr curry

-- | Arrow with uncurry.
uncurryA :: Arrow a => a (b -> c -> d) ((b, c) -> d)
uncurryA = arr uncurry

-- | Arrow with swap.
swapA :: Arrow a => a (b, c) (c, b)
swapA = arr swap

-- | Arrow with liftM2.
liftA2 :: Arrow a => (b -> c -> d) -> a e b -> a e c -> a e d
liftA2 f g h = g &&& h >>> arr (uncurry f)

-- | Arrow with id.
idA :: Arrow a => a b b
idA = arr id

-- | Fst with arrow.
fstA :: Arrow a => a (b, c) b
fstA = arr fst

-- | Snd with arrow.
sndA :: Arrow a => a (b, c) c
sndA = arr snd

-- | Zip with map.
zipMap :: (a -> b, a -> c) -> (a -> (b, c))
zipMap (g, h) s = (g s, h s)

-- | Unzip with map.
unzipMap :: (a -> (b,c)) -> (a -> b, a -> c)
unzipMap h = (fst . h, snd . h)

-- | Trace.
trace :: ((a, c) -> (b, c)) -> a -> b
trace f b = c
    where (c, d) = f (b, d)

-- | Like =~ , but don't care case senstivie.
(=~^) :: (RegexLike Regex source1, RegexMaker Regex CompOption ExecOption source) => source1 -> source -> Bool
source =~^ regex = matchTest (makeRegexOpts (defaultCompOpt {caseSensitive = False}) defaultExecOpt regex) source 

-- | Intersect two file path from end.
fileIntersectEnd :: FilePath -> FilePath -> FilePath
fileIntersectEnd xs ys = 
  -- Remove otiose characters before path separator 
  -- if have previous directory.
  case findIndex (== pathSeparator) zs of
    Just i  -> drop (succ i) zs
    Nothing -> zs
  -- Get common string from end.
  where zs = intersectEnd xs ys 

-- | Get upper directory name.
getUpperDirectoryName :: FilePath -> FilePath
getUpperDirectoryName = takeFileName . dropTrailingPathSeparator . dropFileName

-- | Get upper directory path.
getUpperDirectory :: FilePath -> FilePath
getUpperDirectory = dropFileName . dropTrailingPathSeparator 

-- | Try to expand file path that start with backtick.
-- Just return original one when file path not start with backtick.
expandBacktickFilePath :: FilePath -> IO FilePath
expandBacktickFilePath filepath = 
    if filepath =~ "^~.*" :: Bool
       then expandFilePath filepath
       else return filepath

-- | Expand file name, 
-- including expand backtick and fill blank in filepath.
expandFileName :: FilePath -> IO FilePath
expandFileName = expandBacktickFilePath . fillFilePathBlank

-- | Like `doesDirectoryExist`, but expand file name first.
doesDirectoryExist_ :: FilePath -> IO Bool
doesDirectoryExist_ = (<=<) doesDirectoryExist expandFileName

-- | Like `doesFileExist`, but expand file name first.
doesFileExist_ :: FilePath -> IO Bool
doesFileExist_ = (<=<) doesFileExist expandFileName

-- | Expand file path.
expandFilePath :: FilePath -> IO FilePath
expandFilePath filepath = fmap (head . lines) (readProcess "bash" ["-c", "echo " ++ filepath] [])

-- | Fill blank in filepath with `\\ `.
fillFilePathBlank :: FilePath -> FilePath
fillFilePathBlank filepath = replaceRegex filepath "[^\\][ ]+" "\\ "
    where 
      -- Replace regex match string with target string. 
      replaceRegex string regex target = do
        -- Test match.
        let matchTextList = string =~ regex :: [MatchText String]
        if null matchTextList
           then 
           -- Return origianl string if match nothing.
               string
           else do
             -- Replace regex match stirng with target string. 
             let indexList = map (snd . head . elems) matchTextList
             replaceMatchText indexList target string
      -- Replace match text.
      replaceMatchText indexList target string
          -- Return current string when list null.
          | null indexList 
              = string
          -- Continue replace.
          | otherwise      
              = matchBefore ++ newMatch ++ replaceMatchText newList target matchAfter
          where 
            (matchOffset, matchLength) = head indexList
            (matchBefore, afterString)  = splitAt matchOffset string
            (match, matchAfter) = splitAt matchLength afterString
            newMatch = head match : target
            adjust = matchOffset + matchLength
            newList = map (\(a,b) -> (a - adjust, b)) $ tail indexList

-- | Get directory content, like `getDirectoryContents`, but
-- check directory exist first, if then return Nothing
-- otherwise remove `.` and `..` from result if directory exist.
getDirectoryContents_ :: FilePath -> IO (Maybe [FilePath])
getDirectoryContents_ dir = do
  -- Fill blank first.
  dirName <- expandFileName dir

  -- Get directory content.
  ifM (doesDirectoryExist dirName)
      -- Filter `.` and `..` when target directory exist.
      ((<<<=) Just getDirectoryContents' dirName)
      -- Otherwise return Nothing. 
      (return Nothing)

-- | Get directory content, like `getDirectoryContents`, 
-- but remove `.` and `..` from result if directory exist.
getDirectoryContents' :: FilePath -> IO [FilePath]
getDirectoryContents' = 
    (<<<=) filterDotDirectory getDirectoryContents

-- | Expand file name first.
takeDirectory_ :: FilePath -> IO FilePath
takeDirectory_ = (<<<=) takeDirectory expandFileName

-- | Filter dot directory. 
filterDotDirectory :: [FilePath] -> [FilePath]
filterDotDirectory = filter notDotDirectory

-- | Filter dot directory apply list.
filterDotDirectoryApply :: (a -> FilePath) -> [a] -> [a]
filterDotDirectoryApply =
    filter . (notDotDirectory .)

-- | File not directory.
notDotDirectory :: FilePath -> Bool
notDotDirectory = (`notElem` [".", ".."])

-- | Like takeFileName, but won't return empty when filepath is root path '/'
-- Instead return '/'.          
takeFileNameExceptRoot :: FilePath -> FilePath
takeFileNameExceptRoot filepath 
    | filepath == "/"
      = filepath
    | otherwise 
      = takeFileName filepath

-- | Get file item.
getFileItems :: FilePath -> IO Int
getFileItems filepath = do 
  let path = toUTF8 filepath 
  ifM (doesFileExist_ path)
          -- Return 1 when occur file.
          (return 1)
          (ifM (doesDirectoryExist_ path)
           -- Reutrn sub-itmes number under directory.
           ((<<<=) length getDirectoryContents' path)
           (return 0))

