module Manatee.Utils.Gtk.Gtk where

import Control.Monad
import System.Glib.UTFString
import Data.Maybe
import System.Gnome.VFS hiding (truncate, directoryOpen)
import Control.Exception (handleJust)
import System.IO
import System.Exit
import System.FilePath
import Distribution.Simple.Utils

import Graphics.UI.Gtk hiding (Statusbar, statusbarNew)
import Graphics.UI.Gtk.Entry.Editable
import Graphics.UI.Gtk.Scrolling.ScrolledWindow
import Graphics.UI.Gtk.Abstract.Container
import Graphics.UI.Gtk.Windows.Window
import Graphics.UI.Gtk.Windows.WindowGroup
import Graphics.UI.Gtk.SourceView.SourceLanguage
import Graphics.UI.Gtk.SourceView.SourceLanguageManager
import Graphics.UI.Gtk.ModelView
import Graphics.UI.Gtk.Vte.Vte
import Graphics.UI.Gtk.Vte.Types

import Manatee.Utils.Generic

import qualified System.Gnome.VFS as VFS

-- | Remove all child of container.
containerRemoveAll :: ContainerClass self => self -> IO ()
containerRemoveAll self = containerForeach self $ containerRemove self

-- | Try to packing widget in box.
-- If @child@ have exist parent, do nothing,
-- otherwise, add @child@ to @parent@.
boxTryPack :: (BoxClass parent, WidgetClass child) => parent -> child -> Packing -> Maybe Int -> IO ()
boxTryPack box widget packing order = do
  parent <- widgetGetParent widget
  when (parent == Nothing) $ do
    boxPackStart box widget packing 0
    applyMaybeTM_ (boxReorderChild box widget) order
    
-- | Try to remove child from parent.    
containerTryRemove :: (ContainerClass parent, WidgetClass child) => parent -> child -> IO ()     
containerTryRemove parent widget = do
  hasParent <- widgetGetParent widget
  unless (hasParent == Nothing) $ containerRemove parent widget

-- | Get source language along with file name.
sourceLanguageForFilename :: SourceLanguageManager -> Maybe String -> IO (Maybe String, Maybe SourceLanguage)
sourceLanguageForFilename lm filename = do
    mbLang <- 
        case filename of
          Just f  -> sourceLanguageManagerGuessLanguage lm (Just f) Nothing
          Nothing -> sourceLanguageManagerGuessLanguage lm Nothing (Just "text/x-text")
    case mbLang of
        Nothing -> return (Nothing,Nothing)
        Just lang -> do
            name <- sourceLanguageGetName lang
            return (Just name, Just lang)

-- | Add window with special window group.
windowNewWithWindowGroup :: WindowGroup -> IO Window
windowNewWithWindowGroup windowGroup = do
  window <- windowNew
  windowGroupAddWindow windowGroup window
  return window

-- | Forcus editable and keep status.
editableFocus :: EditableClass self => self -> IO ()
editableFocus entry = do
  (start, end) <- editableGetSelectionBounds entry
  widgetGrabFocus $ castToWidget entry
  editableSelectRegion entry start end
  editableSetPosition entry start
  
-- | Get unselect text.
editableGetUnselectText :: EditableClass self => self -> IO String
editableGetUnselectText ed = do
  allText   <- editableGetAllText ed         -- get entry content
  (start, _) <- editableGetSelectionBounds ed -- get highglitht bound
  return $ take start allText               -- remove highglight completion part from content

-- | Get all text.
editableGetAllText :: EditableClass self => self -> IO String
editableGetAllText ed = editableGetChars ed 0 (-1)

-- | Set text.
editableSetText :: EditableClass self => self -> String -> IO ()
editableSetText ed content = do
  editableDeleteText ed 0 (-1)
  end <- editableInsertText ed content 0
  editableSetPosition ed end

-- | Transform to UTF string.
toUTFString :: String -> IO String
toUTFString= (<=<) readUTFString newUTFString

-- | Get x coordinate of Rectangle.
rectangleX :: Rectangle -> Int
rectangleX (Rectangle x _ _ _) = x

-- | Get y coordinate of Rectangle.
rectangleY :: Rectangle -> Int
rectangleY (Rectangle _ y _ _) = y

-- | Get width of Rectangle.
rectangleWidth :: Rectangle -> Int
rectangleWidth (Rectangle _ _ w _) = w

-- | Get height of Rectangle.
rectangleHeight :: Rectangle -> Int
rectangleHeight (Rectangle _ _ _ h) = h + 2

-- | Apply treeview vadjustment.
treeViewApplyVAdjustment :: TreeViewClass self => self -> (Adjustment -> IO ()) -> IO ()
treeViewApplyVAdjustment treeView f = 
    applyMaybeTM_ f =<< treeViewGetVAdjustment treeView

-- | Move to tree view begin.
treeViewBegin :: TreeViewClass self => self -> IO ()
treeViewBegin treeView = 
    treeViewApplyModel 
    (\_ -> treeViewSetCursor treeView [0] Nothing) treeView

-- | Move to tree view end.
treeViewEnd :: TreeViewClass self => self -> IO ()
treeViewEnd treeView = 
    treeViewApplyVAdjustment treeView 
    (\aj -> treeViewMoveCursorToY treeView 
         =<< (<<<=) pred adjustmentGetUpper aj) 

-- | Get attribute of focus cell.
treeViewGetCellAttr :: TreeViewClass self => (Rectangle -> Int) -> self -> IO Int
treeViewGetCellAttr f treeView = do
  (path, column) <- treeViewGetCursor treeView
  case column of
    -- Get attribute.
    Just c  -> (return . f) =<< treeViewGetCellArea treeView (Just path) c
    -- Otherwise return 0.
    Nothing -> return 0

-- | Get cell x coordinate.
treeViewGetCellX :: TreeViewClass self => self -> IO Int
treeViewGetCellX treeView = do
  x <- treeViewGetCellAttr rectangleX treeView
  (tx,_) <- treeViewWidgetToTreeCoords treeView (x, 0)
  return tx

-- | Get cell y coordinate.
treeViewGetCellY :: TreeViewClass self => self -> IO Int
treeViewGetCellY treeView = do
  y <- treeViewGetCellAttr rectangleY treeView
  (_, ty) <- treeViewWidgetToTreeCoords treeView (0, y)
  return ty

-- | Get cell width.
treeViewGetCellWidth :: TreeViewClass self => self -> IO Int
treeViewGetCellWidth = treeViewGetCellAttr rectangleWidth

-- | Get cell height.
treeViewGetCellHeight :: TreeViewClass self => self -> IO Int
treeViewGetCellHeight = treeViewGetCellAttr rectangleHeight 

-- | Scroll line up.
treeViewScrollLineUp :: TreeViewClass self => self -> IO ()
treeViewScrollLineUp treeView = 
    treeViewScrollVertical treeView 
        =<< (<<<=) integralToDouble treeViewGetCellHeight treeView

-- | Scroll line down.
treeViewScrollLineDown :: TreeViewClass self => self -> IO ()
treeViewScrollLineDown treeView = 
    treeViewScrollVertical treeView 
        =<< (<<<=) (\x -> - (integralToDouble x)) treeViewGetCellHeight treeView

-- | Scroll page up.
treeViewScrollPageUp :: TreeViewClass self => self -> IO ()
treeViewScrollPageUp treeView = 
    treeViewApplyVAdjustment treeView (\aj -> treeViewScrollVertical treeView  
                                           =<< adjustmentGetPageSize aj)

-- | Scroll page down.
treeViewScrollPageDown :: TreeViewClass self => self -> IO ()
treeViewScrollPageDown treeView = 
    treeViewApplyVAdjustment treeView (\aj -> treeViewScrollVertical treeView 
                                           =<< (<<<=) (\x -> - x) adjustmentGetPageSize aj)

-- | Scroll vertical.
treeViewScrollVertical :: TreeViewClass self => self -> Double -> IO ()
treeViewScrollVertical treeView inc =
    treeViewApplyVAdjustment treeView 
    (\aj -> do
       -- Get adjustment arguments.
       ps <- adjustmentGetPageSize aj
       vl <- adjustmentGetValue aj
       ur <- adjustmentGetUpper aj
       lr <- adjustmentGetLower aj

       -- Get y coordinate and height of cell.
       cellY <- (<<<=) integralToDouble treeViewGetCellY treeView
       cellH <- (<<<=) integralToDouble treeViewGetCellHeight treeView

       -- Adjust scroll value.
       let iv = vl + inc
           av = if inc >= 0
                   -- Don't bigger than upper value when scroll up.
                   then min iv $ ur - cellH
                   -- Don't less than lower value when scroll down.
                   else max iv lr
       adjustmentSetValue aj av

       -- Adjust iter postion.
       top <- adjustmentGetValue aj
       let bottom = top + ps
       if top >= cellY
          -- Adjust iter to first line in view visual port 
          -- when iter is up out view visual port.
          then 
              treeViewMoveCursorToY treeView $ cellY + inc
          -- Adjust iter to last line in view visual port
          -- when iter is down out view viesual port.
          else 
              when (bottom <= cellY) $ 
                   treeViewMoveCursorToY treeView $ bottom - 1)

-- | Get next node.
treeViewNextNode :: TreeViewClass self => self -> IO ()
treeViewNextNode treeView = 
    treeViewMoveCursorToY treeView
        =<< (<<<=) integralToDouble (\tv -> liftM2 (+) (treeViewGetCellY tv) (treeViewGetCellHeight tv)) treeView

-- | Get previous node.
treeViewPrevNode :: TreeViewClass self => self -> IO ()
treeViewPrevNode treeView = 
    treeViewMoveCursorToY treeView 
        =<< (<<<=) integralToDouble (\tv -> liftM2 (-) (treeViewGetCellY tv) (treeViewGetCellHeight tv)) treeView

-- | Move cursor to global corrdinate.
treeViewMoveCursorToY :: TreeViewClass self => self -> Double -> IO ()
treeViewMoveCursorToY treeView y = 
    treeViewApplyVAdjustment treeView 
    (\aj -> do
       -- Get current top Y coordinate of visual window.
       vl <- adjustmentGetValue aj
       -- Scroll down to make target cursor position in visual window,
       -- otherwise, treeViewGetPathAtPos can't work.
       when (vl >= y) $ treeViewScrollToPoint treeView 0 (truncate y)
       -- Move to point.
       -- 
       -- FIXME: when (vl >= y) and just have one row display at top of treeView
       -- should just move focus to this row and don't scroll page for smart display
       -- but function `treeViewSetCursor` will scroll page after focus cursor, 
       -- we should develop `treeViewSetCursor_` just focus path and don't scroll page.
       treeViewApplyPointPath treeView 0 y (\(path, _, _) -> treeViewSetCursor treeView path Nothing))

-- | Focus to current cursor postion.
treeViewFocus :: TreeViewClass self => self -> IO ()
treeViewFocus treeView =
    treeViewMoveCursorToY treeView  
        =<< (<<<=) integralToDouble treeViewGetCellY treeView

-- | Apply treeView model.
treeViewApplyModel :: TreeViewClass self => (TreeModel -> IO ()) -> self -> IO ()
treeViewApplyModel f treeView =
    applyMaybeTM_ f =<< treeViewGetModel treeView

-- | Apply cell path.
treeViewApplyCellPath :: TreeViewClass self => self -> ((TreePath, TreeViewColumn, Point) -> IO ()) -> IO ()
treeViewApplyCellPath treeView f = do
  x <- (<<<=) integralToDouble treeViewGetCellX treeView
  y <- (<<<=) integralToDouble treeViewGetCellY treeView
  treeViewApplyPointPath treeView x y f

-- | Apply point path.
treeViewApplyPointPath :: TreeViewClass self => self -> Double -> Double -> ((TreePath, TreeViewColumn, Point) -> IO ()) -> IO ()
treeViewApplyPointPath treeView x y f = do
  -- Transform global coordinate to local coordinate.
  p <- treeViewTreeToWidgetCoords treeView (truncate x, truncate y)

  -- Function treeViewGetPathAtPos just work with local position 
  -- that above 0, if any coordinate below 0, can't work.
  applyMaybeTM_ f =<< treeViewGetPathAtPos treeView p

-- | Remove all coluns from treeView.
treeViewRemoveColumns :: TreeViewClass self => self -> IO ()
treeViewRemoveColumns treeView = 
    mapM_ (treeViewRemoveColumn treeView) 
        =<< treeViewGetColumns treeView

treeModelSortGetRow :: (TreeModelSortClass self, 
                       TypedTreeModelClass model) =>
                      model row
                    -> self
                    -> TreeIter
                    -> IO row
treeModelSortGetRow model = 
    (<=<) (treeModelGetRow model) . treeModelSortConvertIterToChildIter

treeViewAddColumnWithTitle :: TreeViewClass self => self -> String -> SortColumnId -> IO TreeViewColumn
treeViewAddColumnWithTitle treeView title sortId = do
  -- Get treeViewColumn.
  tvc <- treeViewColumnNew
  treeViewAppendColumn treeView tvc

  -- Set attribute.
  set tvc [treeViewColumnTitle          := title
          ,treeViewColumnResizable      := True
          ,treeViewColumnSortIndicator  := True
          ,treeViewColumnSortColumnId   := sortId]

  return tvc

-- | Handle VFS error.
handleVFSError :: IO a -> IO a
handleVFSError = 
    handleJust VFS.errors (\(VFS.Error error)  -> do
                             hPutStrLn stderr $ "VFS error: " ++ show error
                             exitFailure)

-- | Init VFS.
initVFS :: IO ()
initVFS =
    VFS.init >>= \success -> unless success $ do 
                              hPutStrLn stderr "could not initialize GnomeVFS"
                              exitFailure

-- | Add trailing with file type.
addTrailingPathSeparatorWithFileType :: FilePath -> FileType -> FilePath
addTrailingPathSeparatorWithFileType fPath fType
    | fType == FileTypeDirectory 
        = addTrailingPathSeparator fPath
    | otherwise
        = fPath

-- | Compare file name with file type.
-- Directory first, then sort file. 
-- All sort is alpha sort.
compareFile :: (FilePath, FileType) -> (FilePath, FileType) -> Ordering
compareFile (fPath1, fType1) (fPath2, fType2) 
    | fType1 == fType2 
      = compare fName1 fName2
    | fType1 == FileTypeDirectory
      = LT
    | fType2 == FileTypeDirectory
      = GT
    | otherwise 
      = compare fName1 fName2
    -- Don't care file case.
    where fName1 = lowercase fPath1        
          fName2 = lowercase fPath2

-- | Scroll line up.
terminalScrollLineUp :: (TerminalClass term, ScrolledWindowClass swc) => term -> swc -> IO ()
terminalScrollLineUp term swc = 
  terminalScrollVertical term swc 
      =<< (<<<=) integralToDouble terminalGetCharHeight term

-- | Scroll line up.
terminalScrollLineDown :: (TerminalClass term, ScrolledWindowClass swc) => term -> swc -> IO ()
terminalScrollLineDown term swc = do
  -- Get font height.
  fontHeight <- (<<<=) integralToDouble terminalGetCharHeight term
  terminalScrollVertical term swc (-fontHeight)

-- | Scroll page up.
terminalScrollPageUp :: (TerminalClass term, ScrolledWindowClass swc) => term -> swc -> IO ()
terminalScrollPageUp term swc = do
  -- Get font height.
  fontHeight <- (<<<=) integralToDouble terminalGetCharHeight term
  terminalScrollVertical term swc (50 * fontHeight)

-- | Scroll page up.
terminalScrollPageDown :: (TerminalClass term, ScrolledWindowClass swc) => term -> swc -> IO ()
terminalScrollPageDown term swc = do
  -- Get font height.
  fontHeight <- (<<<=) integralToDouble terminalGetCharHeight term
  terminalScrollVertical term swc (- (50 * fontHeight))

-- | Scroll vertical of terminal widget.
terminalScrollVertical :: (TerminalClass term, ScrolledWindowClass swc) => term -> swc -> Double -> IO ()
terminalScrollVertical term swc inc = do
  -- Get adjustment arguments.
  aj <- scrolledWindowGetVAdjustment swc
  vl <- adjustmentGetValue aj
  ur <- adjustmentGetUpper aj
  lr <- adjustmentGetLower aj
  
  -- Get font height.
  fontHeight <- (<<<=) integralToDouble terminalGetCharHeight term
  
  -- Adjust scroll value.
  let iv = vl + inc
      av = if inc >= 0
              -- Don't bigger than upper value when scroll up.
              then min iv (ur - fontHeight)
              -- Don't less than lower value when scroll down.
              else max iv lr
  adjustmentSetValue aj av

-- | Move to begin of terminal.
terminalBegin :: ScrolledWindowClass swc => swc -> IO ()  
terminalBegin swc = do
  -- Get adjustment arguments.
  aj <- scrolledWindowGetVAdjustment swc
  
  adjustmentSetValue aj 0

-- | Move to begin of terminal.
terminalEnd :: ScrolledWindowClass swc => swc -> IO ()  
terminalEnd swc = do
  -- Get adjustment arguments.
  aj <- scrolledWindowGetVAdjustment swc
  ps <- adjustmentGetPageSize aj
  ur <- adjustmentGetUpper aj
  
  adjustmentSetValue aj (ur - ps)

